#!/bin/bash

SCPATH=~/sshTools ## Path to SSH Tools. No trailing slash. This can also be set in the configuration file.

## Part of Arxx SSH Tools. Published at: <https://arxx.se>.
## Written by: Johan Hyllstedt. You can reach me at: <code@arxx.se>.
## See README for more information.

## Copyright 2010, 2011, 2012 Johan Hyllstedt.

## This file is part of Arxx SSH Tools.

## Arxx SSH Tools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## Arxx SSH Tools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with Arxx SSH Tools.  If not, see <http://www.gnu.org/licenses/>.

function printIC
{
  ## Print Information & Continues, but only if vebose

  if [ $VERBOSE -eq 1 ] && [ $SILENT -eq 0 ];
  then
    echo -e "I: ${1:-"Unknown Information"}"
  fi
}

function printEE
{
  ## Print Error & Exit
  ## Prints to stdout & stderr and exits

  echo -e "## E: ${1:-"Unknown Error"}" 1>&2

  exit 1
}

function mkConf
{
  if [ -f "$PREFS" ];
  then
    printEE "The file '$PREFS' already exists, I will not overwrite this!"
  fi

  echo "I: Making sample configuration file in: '$PREFS'."

  cat >"$PREFS"<<EOF
## Sample configuration file for Arxx SSH Tools.
## The values set here at the beginning are the default values used by the tools.
## Note that most values here can be overridden at runtime by specifying other values in the input parameters.
## Comments like this are ignored.

## You can also set the path to the tools here. No trailing slash.
SCPATH=~/sshTools

## To always use SILENT mode by default set this to 1.
SILENT=0
SSILENT=0         ## Also set this to 1 for super-silent mode (-ss).

## To always use VERBOSE mode by default set this to 1 (disable SILENT above first).
VERBOSE=0

## To always use pad by default set this to 1.
ENABLEPAD=0

## To use a specific pad length set this to [NR]d, so for 5 it will be "5d".
PAD="2d"

## To always use only SSH keys by default set this to 1.
KEYSONLY=0

## To always use HOSTMODE per default set this to 1.
HOSTMODE=0

## To specify a default password you can do this here, note that KEYSONLY must be 0.
## This naturally also sets password mode as the default.
## Also note that anyone who can read this file can also get the password of course!
SSHPASSWD=""

## To use CRITICALERROR mode by default set this to 1.
CRITICALERROR=0

## To use EXITERROR mode by default set this to 1.
EXITERROR=0

## To always use FILEMODE per default and load hosts from file, set these below.
FILEMODE=0
FILE=""
EOF

  exit 0
}

function mkPw
{
  if [ -f "$PWPATH" ];
  then
    printEE "The file '$PWPATH' already exists, I will not overwrite this!"
  fi

  echo "I: Making sample configuration file in: '$PWPATH'."

  cat >"$PWPATH"<<EOF
## Sample password file for Arxx SSH Tools wrapper scripts.
## Note that most values here can be overridden at runtime by specifying other values in the input parameters.
## Comments like this are ignored.

## You can use any slot number you like between 1-Unlimited, but you must match the INPUTLEN to the amount of digits your slots will use.
## Also keep in mind that slot 6 is reserved for the "Ask password" option, so you cannot use that slot.

## Description of pw
DESC[1]="Password 1"
DESC[2]="Password for myuser"
DESC[3]="Pw will be copied to clipboard"
DESC[6]="" ## Enter pass

## Format is: USERNAME PASSWD [NR] - WHITESPACE is the delimiter.
## A 1 in the third field also copies the password to clipboard, entering nothing here disables this feature (make sure to set CLEARCLIP if enabled).
## A 2 in the third field assumes you first want to sudo to root using the password you choose. NOTE: this is not supported yet with SCP!
## A 3 assume you want both above.
USPA[1]="root pAssword"
USPA[2]="myuser PASSwoord"
USPA[3]="myuser Clipword 1"
USPA[6]="" ## Enter pass

## This should be 1 for slots 1-9, 2 for 1-99 and so on.
INPUTLEN=1

## Set to 1 clears clipboard on exit.
CLEARCLIP=0

## Clear clipboard only if pw was copied to clipboard
CLIPUSECLEAR=1

## What username use if not specified.
USERN="root"

## To always sudo to root first set this to 1.
ALWAYSSUDO=0

## Remember which password used for hosts automatically and use it nexttime.
REMEMBERPW=0
EOF

  exit 0
}

function askPass
{
  ## Asks the user to select a password and sets SSHPASSWD accordingly.

  USERN="$ORIGUSER"
  SSHPASSWD=""
  CLIP=""
  c=1
  REMPW_MATCHED=0
  USEPWFROMFILE=0
  unset data

  if [ "$REMEMBERPW" == "1" ] && [ $FIRSTASK -eq 1 ];
  then
    REMPW="`grep -E "^$HOSTNAMEORIP:" "$PWNRPATH" | tail -n1`"

    if [ -n "$REMPW" ];
    then
      ## Matched
      REMPW_MATCHED=1

      if [ "$USEPWSET" != "1" ];
      then
        USEPW="`echo "$REMPW" | awk -F ':' '{print $2}'`"

        if [ "$USEPW" == "6" ];
        then
          PWFROMFILE="`echo "$REMPW" | awk -F ':' '{print $3}'`"
          USEPWFROMFILE=1
        else
          USEPWFROMFILE=0
        fi
      fi
    fi
  elif [ "$USEPWSET" != "1" ] && [ $FIRSTASK -eq 0 ] && [ "$REMEMBERPW" == "1" ];
  then
    USEPW=""
  fi

  if [ -z "$USEPW" ] && [ $REMPW_MATCHED -eq 0 ];
  then
    for i in "${DESC[@]}"
    do
      if [ "$c" == "6" ];
      then
        echo -n "6: Enter password. "
      else
        echo -n "$c: $i. "
      fi

      c=$((c+1))
    done

    echo -n "*: Quit. C? "
    read -n $INPUTLEN ANSW
    echo
  else
    ANSW="$USEPW"
  fi

  c=1
  MATCH=0

  if [ "$ANSW" == "6" ];
  then
    if [ $USEPWFROMFILE -eq 1 ];
    then
      SSHPASSWD="$PWFROMFILE"
    else
      echo -n "Password: "
      read -s SSHPASSWD
      echo
    fi
    MATCH=1
  else
    for i in "${USPA[@]}"
    do
      if [ "$ANSW" == "$c" ];
      then
        data=($i)
        USERN=${data[0]}
        SSHPASSWD=${data[1]}
        CLIP=${data[2]}
        MATCH=1
        break
      fi

      c=$((c+1))
    done
  fi

  if [ $CHANGEPW -eq 1 ] && [ $MATCH -eq 1 ];
  then
    ## Remove all previous occurances and set a new one
    sed -i "/^${HOSTNAMEORIP}:/d" $PWNRPATH

    if [ "$ANSW" == "6" ];
    then
      echo "$HOSTNAMEORIP:$ANSW:$SSHPASSWD" >>$PWNRPATH
    else
      echo "$HOSTNAMEORIP:$ANSW" >>$PWNRPATH
    fi
  fi

  if [ "$REMEMBERPW" == "1" ];
  then
    if [ $REMPW_MATCHED -eq 0 ] && [ $MATCH -eq 1 ] && [ $FIRSTASK -eq 1 ];
    then
      ## Did not use saved pw, and we matched, save it

      if [ "$ANSW" == "6" ];
      then
        echo "$HOSTNAMEORIP:$ANSW:$SSHPASSWD" >>$PWNRPATH
      else
        echo "$HOSTNAMEORIP:$ANSW" >>$PWNRPATH
      fi
    fi
  fi

  USESUDO=0 ## Always reset it

  if [ "$CLIP" == "1" ];
  then
    pwToClip
  elif [ "$CLIP" == "2" ];
  then
    USESUDO=1
    echo "W: Unsupported mode. This might not work!" ## SCP specific
  elif [ "$CLIP" == "3" ];
  then
    USESUDO=1
    echo "W: Unsupported mode. This might not work!" ## SCP specfic

    pwToClip
  fi

  if [ "$ALWAYSSUDO" == "1" ];
  then
    USESUDO=1
  fi

  if [ $FIRSTASK -eq 1 ];
  then
    FIRSTASK=0
  fi

  if [ $MATCH -eq 0 ];
  then
    echo "Bye."

    exit 0
  fi
}

function pwToClip
{
  ## Copies password to clipboard.

  echo -n "$SSHPASSWD" | xclip

  if [ $CLIPUSECLEAR -eq 1 ];
  then
    export CLEARCLIP=1
  fi
}

function checkHost
{
  ## Checks if first arg is in the hostfile or available via DNS. If not in either sends a warning message.
  ## If optional arg 2 says "1" then additional formatting is enabled and input should be in the form user@host:path -> host

  if [ -z "$1" ];
  then
    return 0
  fi

  if [ "$2" == "1" ];
  then
    M="`echo "$1" | sed -n 's/.*@//p' | sed -n 's/:.*//p'`"

    if [ -z "$M" ];
    then
      M="`echo "$1" | sed -n 's/:.*//p'`"

      if [ -z "$M" ];
      then
        echo "W: Failed to parse hostname: $1."
        return
      fi
    fi
  else
    M="$1"
  fi

  ## Check hosts file
  RESULTFI="`grep -si "$M" /etc/hosts`"

  ## Dns lookup for 1 sec
  host -W 1 "$M" 1>/dev/null 2>/dev/null

  RESULTDNS=$?

  if [ -z "$RESULTFI" ] && [ $RESULTDNS -gt 0 ];
  then
    ## Both checks are bad.
    echo "W: Hostname: $M, not found."
  fi
}

function mod_fix
{
  if [ -z "`echo "$1" | grep -soE '\@'`" ];
  then
    MOD="$USERN@$1"
  else
    MOD="$1"
  fi
}

## Global arg
GETNEXT=0
C=0
AUTO=0
VERBOSE=0
SILENT=0
PREFS=~/.sshTools.cfg
SYSPREFS="/etc/ast.cfg"
USEPW=""
INPUTLEN=1
CLEARCLIP=0
USERN="`whoami`" ## To mimic default ssh behaviour
PWPATH=~/.pws.cfg
PWNRPATH=~/.pws-nrs.cfg
USESUDO=0
ALWAYSSUDO=0
SCPSCRIPT="ScpSinglePasswd.exp"
REMEMBERPW=0
CHANGEPW=0
NOREMPW=0
FIRSTASK=1
USEPWSET=0

## Easy/lazy arg handling
for ARG in "$@"
do
  if [ $GETNEXT -eq 1 ];
  then
    GETNEXT=0

    USEPW="$ARG"
    USEPWSET=1
    AUTO=1

    continue
  fi

  if [ "$ARG" == "-p" ];
  then
    GETNEXT=1
  elif [ "$ARG" == "-v" ] || [ "$ARG" == "--verbose" ];
  then
    VERBOSE=1
  elif [ "$ARG" == "-e" ] || [ "$ARG" == "--change-pw" ];
  then
    CHANGEPW=1
  elif [ "$ARG" == "-u" ] || [ "$ARG" == "--use-saved-pw" ];
  then
    REMEMBERPW=1
  elif [ "$ARG" == "-n" ] || [ "$ARG" == "--no-saved-pw" ];
  then
    NOREMPW=1
  elif [ "$ARG" == "-S" ] || [ "$ARG" == "-r" ] || [ "$ARG" == "--save-pw" ];
  then
    REMEMBERPW=1
  elif [ "$ARG" == "--mk-conf" ];
  then
    mkConf
  elif [ "$ARG" == "--mk-pwf" ];
  then
    mkPw
  elif [ "$ARG" == "-h" ] || [ "$ARG" == "--help" ];
  then
    echo "SCP Wrapper - Arxx SSH Tools <https://arxx.se>."
    echo
    echo "Syntax: $0 [ OPTION ] .. SRC [ SRC ] .. DEST"
    echo
    echo "Examples:"
    echo "  $ $0 root@172.16.1.1:/tmp/test test/"
    echo "  $ $0 -p 3 file1 file2 files* user1@remotehost:remotefolder/"
    echo "  $ $0 user1@remotehost:remotefolder/ root@remotehost2:remotefolder/"
    echo
    echo "It is possible to copy neatly from src-host to dest-host, without the normal hassles."
    echo "You can also specify which pw to use with: -p PASSWORD_NUMBER."
    echo
    echo "Commandline options:"
    echo "  --mk-conf          - Make sample Arxx SSH Tools config file."
    echo "  --mk-pwf           - Make sample password file."
    echo "  -v, --verbose      - Verbose mode."
    echo "  -p PWNR            - Connect with password number PWNR."
    echo "  -r, -S, --save-pw  - Save password this time only (disregard config setting)."
    echo "  -u, --use-saved-pw - Use a saved password this time only (disregard config setting)."
    echo "  -n, --no-saved-pw  - Do not use any saved passwords and do not save this time."
    echo "  -e, --change-pw    - Change saved password for a saved host this time only."

    exit 0
  else
    C=$((C+1))
    ARGS[$C]="$ARG"
  fi
done

if [ "$REMEMBERPW"  == "1" ];
then
  REMPWSET=1
else
  REMPWSET=0
fi

if [ -z "${ARGS[1]}" ];
then
  echo "What? Use -h!"
  exit 1
fi

if [ -r "$SYSPREFS" ];
then
  printIC "Using system preferences set in: $SYSPREFS."

  . "$SYSPREFS"
fi

if [ -r "$PREFS" ];
then
  printIC "Using preferences set in: $PREFS."

  . "$PREFS"
else
  printIC "Config file not found ($PREFS) so using defaults or system config instead."
fi

if [ ! -r "$PWPATH" ];
then
  printEE "Cannot read pw file: $PWPATH."
else
  . "$PWPATH"
fi

if [ $NOREMPW -eq 1 ] || [ $CHANGEPW -eq 1 ];
then
  REMEMBERPW=0
fi

if [ "$REMPWSET"  == "1" ];
then
  REMEMBERPW=1
fi

## Set this here.
ORIGUSER="$USERN" ## Dup

if [ -z "`echo "${ARGS[1]}" | grep -soE '\@'`" ];
then
  START="$USERN@${ARGS[1]}"
else
  START="${ARGS[1]}"
fi

RETVAL=1
USENSRC=0
if [ -n "${ARGS[3]}" ];
then
  ## We have a pack of files
  len="${#ARGS[@]}"
  SRC=""
  DEST="${ARGS[len]}"
  ARGS[len]=""

  for f in ${ARGS[@]}
  do
    if [ "$SRC" == "" ];
    then
      SRC="$f"
    else
      SRC="${SRC}\n${f}"
    fi
  done

  ## Array trickery...
  SRC="`echo "$SRC" | sed 's/\\\n/\n/g'`"

  NSRC="`echo "$SRC" | sed 's/ /___WHITESPACE___/g'`"
  NSRC="`echo "$NSRC" | sed 's/\n/ /g'`"
  NSRC=($NSRC)
  LEN=${#NSRC[@]}

  for ((i = 0; i < LEN; i++ ))
  do
    NSRC[$i]="`echo "${NSRC[$i]}" | sed 's/___WHITESPACE___/ /g'`"
  done
  USENSRC=1
else
  SRC="${ARGS[1]}"
  DEST="${ARGS[2]}"
fi
DIRECTION=0 ## 1: SRC 2: DEST
STOP=0

if [ -n "`echo "$SRC" | grep -soE ":"`" ] && [ -z "`echo "$DEST" | grep -soE ":"`" ];
then
  DIRECTION=1
  DIR="${SRC[@]}"
elif [ -n "`echo "$DEST" | grep -soE ":"`" ] && [ -z "`echo "$SRC" | grep -soE ":"`" ];
then
  DIRECTION=2
  DIR="$DEST"
elif [ -n "`echo "$DEST" | grep -soE ":"`" ] && [ -n "`echo "$SRC" | grep -soE ":"`" ];
then
  ## From remotehost to remotehost, tricky one, orig scp does this quite stupid and also it has its reasons for this so lets not be mad.
  ## However we have a slightly better way, which again is not perfect...

  DIRECTION=3
  DIR=""
else
  printEE "Cannot figure out which direction you are copying files, local copy?"
fi

while [ $RETVAL -ne 0 ] && [ $STOP -eq 0 ];
do
  breaked=0
  MULTI=0
  if [ $USENSRC -eq 1 ];
  then
    DIR="$DEST"
    MULTI=1
  fi

  checkHost "$DIR" 1

  if [ $DIRECTION -eq 3 ];
  then
    echo "## I: First copying remote file to local tempdir. Choose login for: $SRC."
  fi

  if [ $DIRECTION -eq 3 ];
  then
    ## We are gonna do two runs, so first lets have the src here.
    DIR="$SRC"
  fi

  HOSTNAMEORIP="`echo "$DIR" | awk -F ':' '{print $1}'`"

  askPass

  if [ "$USESUDO" == "1" ];
  then
    #SCPSCRIPT="ScpSinglePasswdSudo.exp"
    SCPSCRIPT="ScpSinglePasswd.exp"
  else
    SCPSCRIPT="ScpSinglePasswd.exp"
  fi

  if [ $MULTI -eq 0 ];
  then
    mod_fix "$DIR" ## sets MOD
  fi

  if [ $DIRECTION -eq 1 ];
  then
    $SCPATH/$SCPSCRIPT "$MOD" "$DEST" "$SSHPASSWD"
  elif [ $DIRECTION -eq 2 ];
  then
    if [ $MULTI -eq 1 ];
    then
      mod_fix "$DEST" ## sets MOD
      for D in ${NSRC[@]}
      do
        $SCPATH/$SCPSCRIPT "$D" "$MOD" "$SSHPASSWD"
        THISRET=$?

        if [ $THISRET -ne 0 ];
        then
          RETVAL=$THISRET
          breaked=1
          break
        fi
      done
    else
      $SCPATH/$SCPSCRIPT "$SRC" "$MOD" "$SSHPASSWD"
    fi
  elif [ $DIRECTION -eq 3 ];
  then
    ## What needs to be done:
    ## 1: Download file to tmp dir umasked 077
    ## 2: Copy that file to remotehost2
    ## 3: Remove tmpdir

    if [ -z "`echo "$SRC" | grep -soE '\@'`" ];
    then
      MOD1="$USERN@$SRC"
    else
      MOD1="$SRC"
    fi

    SSHPASSWD1="$SSHPASSWD" ## Holds first pass
    echo "## I: Now choose login for: $DEST."

    ## Since this is our second pass, maybe we do not care.
    FIRSTASK=1

    HOSTNAMEORIP="`echo "$DEST" | awk -F ':' '{print $1}'`"

    askPass
    SSHPASSWD2="$SSHPASSWD" ## Now holds second

    if [ "$USESUDO" == "1" ];
    then
      #SCPSCRIPT="ScpSinglePasswdSudo.exp"
      SCPSCRIPT="ScpSinglePasswd.exp"
    else
      SCPSCRIPT="ScpSinglePasswd.exp"
    fi

    if [ -z "`echo "$DEST" | grep -soE '\@'`" ];
    then
      MOD2="$USERN@$DEST"
    else
      MOD2="$DEST"
    fi

    umask 077
    TDIR="${TMPDIR:-"/tmp"}/scptemp_${RANDOM}$$/" ## Honor export temp dir
    mkdir "$TDIR" || printEE "Failed to make tmp dir: $TDIR."
    $SCPATH/$SCPSCRIPT "$MOD1" "$TDIR" "$SSHPASSWD1" || printEE "Failed to copy file to local temp dir."

    TMPFILES="`ls $TDIR* | xargs`"

    $SCPATH/$SCPSCRIPT "$TMPFILES" "$MOD2" "$SSHPASSWD2"

    rm -r "$TDIR"
  fi

  if [ $breaked -eq 0 ];
  then
    RETVAL=$?
  fi

  if [ $AUTO -eq 1 ];
  then
    ## Stop after first attempt
    STOP=1
  fi
done

echo "------------------------------------------------------------"
echo "## I: Done!"

if [ ${CLEARCLIP} -eq 1 ];
then
  echo -n "CLEARED" | xclip
fi
