#!/bin/bash

SCPATH=~/sshTools ## Path to SSH Tools. No trailing slash. This can also be set in the configuration file.

## Part of Arxx SSH Tools. Published at: <https://arxx.se>.
## Written by: Johan Hyllstedt. You can reach me at: <code@arxx.se>.
## See README for more information.

## Copyright 2010, 2011, 2012 Johan Hyllstedt.

## This file is part of Arxx SSH Tools.

## Arxx SSH Tools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## Arxx SSH Tools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with Arxx SSH Tools.  If not, see <http://www.gnu.org/licenses/>.

function printIC
{
  ## Print Information & Continues, but only if vebose

  if [ $VERBOSE -eq 1 ] && [ $SILENT -eq 0 ];
  then
    echo -e "I: ${1:-"Unknown Information"}"
  fi
}

function printEE
{
  ## Print Error & Exit
  ## Prints to stdout & stderr and exits

  echo -e "## E: ${1:-"Unknown Error"}" 1>&2

  exit 1
}

function mkConf
{
  if [ -f "$PREFS" ];
  then
    printEE "The file '$PREFS' already exists, I will not overwrite this!"
  fi

  echo "I: Making sample configuration file in: '$PREFS'."

  cat >"$PREFS"<<EOF
## Sample configuration file for Arxx SSH Tools.
## The values set here at the beginning are the default values used by the tools.
## Note that most values here can be overridden at runtime by specifying other values in the input parameters.
## Comments like this are ignored.

## You can also set the path to the tools here. No trailing slash.
SCPATH=~/sshTools

## To always use SILENT mode by default set this to 1.
SILENT=0
SSILENT=0         ## Also set this to 1 for super-silent mode (-ss).

## To always use VERBOSE mode by default set this to 1 (disable SILENT above first).
VERBOSE=0

## To always use pad by default set this to 1.
ENABLEPAD=0

## To use a specific pad length set this to [NR]d, so for 5 it will be "5d".
PAD="2d"

## To always use only SSH keys by default set this to 1.
KEYSONLY=0

## To always use HOSTMODE per default set this to 1.
HOSTMODE=0

## To specify a default password you can do this here, note that KEYSONLY must be 0.
## This naturally also sets password mode as the default.
## Also note that anyone who can read this file can also get the password of course!
SSHPASSWD=""

## To use CRITICALERROR mode by default set this to 1.
CRITICALERROR=0

## To use EXITERROR mode by default set this to 1.
EXITERROR=0

## To always use FILEMODE per default and load hosts from file, set these below.
FILEMODE=0
FILE=""
EOF

  exit 0
}

function mkPw
{
  if [ -f "$PWPATH" ];
  then
    printEE "The file '$PWPATH' already exists, I will not overwrite this!"
  fi

  echo "I: Making sample configuration file in: '$PWPATH'."

  cat >"$PWPATH"<<EOF
## Sample password file for Arxx SSH Tools wrapper scripts.
## Note that most values here can be overridden at runtime by specifying other values in the input parameters.
## Comments like this are ignored.

## You can use any slot number you like between 1-Unlimited, but you must match the INPUTLEN to the amount of digits your slots will use.
## Also keep in mind that slot 6 is reserved for the "Ask password" option, so you cannot use that slot.

## Description of pw
DESC[1]="Password 1"
DESC[2]="Password for myuser"
DESC[3]="Pw will be copied to clipboard"
DESC[6]="" ## Enter pass

## Format is: USERNAME PASSWD [NR] - WHITESPACE is the delimiter.
## A 1 in the third field also copies the password to clipboard, entering nothing here disables this feature (make sure to set CLEARCLIP if enabled).
## A 2 in the third field assumes you first want to sudo to root using the password you choose. NOTE: this is not supported yet with SCP!
## A 3 assume you want both above.
USPA[1]="root pAssword"
USPA[2]="myuser PASSwoord"
USPA[3]="myuser Clipword 1"
USPA[6]="" ## Enter pass

## Dirs, format: REGEXP DIR.
## This will let you automatically cd to a specifed directory after loggin in. The regexp is matched against the hostname/ip you are connecting to.
## Keep in mind that many posts slows the login proccess down.
DIRS[1]="127.0.0.* /tmp"
DIRS[2]=".*15.2.* /tmp/bigfiles"

## This should be 1 for slots 1-9, 2 for 1-99 and so on.
INPUTLEN=1

## Set to 1 clears clipboard on exit.
CLEARCLIP=0

## Clear clipboard only if pw was copied to clipboard
CLIPUSECLEAR=1

## What username use if not specified.
USERN="root"

## To always sudo to root first set this to 1.
ALWAYSSUDO=0

## Remember which password used for hosts automatically and use it nexttime.
REMEMBERPW=0
EOF

  exit 0
}

function askPass
{
  ## Asks the user to select a password and sets SSHPASSWD accordingly.

  USERN="$ORIGUSER"
  SSHPASSWD=""
  CLIP=""
  c=1
  REMPW_MATCHED=0
  USEPWFROMFILE=0
  unset data

  if [ "$REMEMBERPW" == "1" ] && [ $FIRSTASK -eq 1 ];
  then
    REMPW="`grep -E "^$HOSTNAMEORIP:" "$PWNRPATH" | tail -n1`"

    if [ -n "$REMPW" ];
    then
      ## Matched
      REMPW_MATCHED=1

      if [ "$USEPWSET" != "1" ];
      then
        USEPW="`echo "$REMPW" | awk -F ':' '{print $2}'`"

        if [ "$USEPW" == "6" ];
        then
          PWFROMFILE="`echo "$REMPW" | awk -F ':' '{print $3}'`"
          USEPWFROMFILE=1
        else
          USEPWFROMFILE=0
        fi
      fi
    fi
  elif [ "$USEPWSET" != "1" ] && [ $FIRSTASK -eq 0 ] && [ "$REMEMBERPW" == "1" ];
  then
    USEPW=""
  fi

  if [ -z "$USEPW" ] && [ $REMPW_MATCHED -eq 0 ];
  then
    for i in "${DESC[@]}"
    do
      if [ "$c" == "6" ];
      then
        echo -n "6: Enter password. "
      else
        echo -n "$c: $i. "
      fi

      c=$((c+1))
    done

    echo -n "*: Quit. C? "
    read -n $INPUTLEN ANSW
    echo
  else
    ANSW="$USEPW"
  fi

  c=1
  MATCH=0

  if [ "$ANSW" == "6" ];
  then
    if [ $USEPWFROMFILE -eq 1 ];
    then
      SSHPASSWD="$PWFROMFILE"
    else
      echo -n "Password: "
      read -s SSHPASSWD
      echo
    fi
    MATCH=1
  else
    for i in "${USPA[@]}"
    do
      if [ "$ANSW" == "$c" ];
      then
        data=($i)
        USERN=${data[0]}
        SSHPASSWD=${data[1]}
        CLIP=${data[2]}
        MATCH=1
        break
      fi

      c=$((c+1))
    done
  fi

  if [ $CHANGEPW -eq 1 ] && [ $MATCH -eq 1 ];
  then
    ## Remove all previous occurances and set a new one
    sed -i "/^${HOSTNAMEORIP}:/d" $PWNRPATH

    if [ "$ANSW" == "6" ];
    then
      echo "$HOSTNAMEORIP:$ANSW:$SSHPASSWD" >>$PWNRPATH
    else
      echo "$HOSTNAMEORIP:$ANSW" >>$PWNRPATH
    fi
  fi

  if [ "$REMEMBERPW" == "1" ];
  then
    if [ $REMPW_MATCHED -eq 0 ] && [ $MATCH -eq 1 ] && [ $FIRSTASK -eq 1 ];
    then
      ## Did not use saved pw, and we matched, save it

      if [ "$ANSW" == "6" ];
      then
        echo "$HOSTNAMEORIP:$ANSW:$SSHPASSWD" >>$PWNRPATH
      else
        echo "$HOSTNAMEORIP:$ANSW" >>$PWNRPATH
      fi
    fi
  fi

  USESUDO=0 ## Always reset it

  if [ "$CLIP" == "1" ];
  then
    pwToClip
  elif [ "$CLIP" == "2" ];
  then
    USESUDO=1
  elif [ "$CLIP" == "3" ];
  then
    USESUDO=1

    pwToClip
  fi

  if [ "$ALWAYSSUDO" == "1" ];
  then
    USESUDO=1
  fi

  if [ $FIRSTASK -eq 1 ];
  then
    FIRSTASK=0
  fi

  if [ $MATCH -eq 0 ];
  then
    echo "Bye."

    exit 0
  fi
}

function dirCheck
{
  ## Arg1 = hostname to connect to.
  ## Checks if host/ip matches a regex and sets return variable NEWDIR accordingly.
  NEWDIR="" ## Empty first
  unset data
  THEHOST="$1"

  for DIR in "${DIRS[@]}"
  do
    data=($DIR)
    REGEX=${data[0]}
    THEDIR=${data[1]}

    if [ "`echo "$THEHOST" | grep -sE "$REGEX"`" != "" ];
    then
      NEWDIR="$THEDIR"
      break
    fi
  done
}

function pwToClip
{
  ## Copies password to clipboard.

  echo -n "$SSHPASSWD" | xclip

  if [ $CLIPUSECLEAR -eq 1 ];
  then
    export CLEARCLIP=1
  fi
}

function checkHost
{
  ## Checks if first arg is in the hostfile or available via DNS. If not in either sends a warning message.

  if [ -z $1 ];
  then
    return 0
  fi

  ## Check hosts file
  RESULTFI="`grep -si "$1" /etc/hosts`"

  ## Dns lookup for 1 sec
  host -W 1 "$1" 1>/dev/null 2>/dev/null

  RESULTDNS=$?

  if [ -z "$RESULTFI" ] && [ $RESULTDNS -gt 0 ];
  then
    ## Both checks are bad.
    echo "W: Hostname not found."
  fi
}

## Global arg
GETNEXT=0
C=0
AUTO=0
VERBOSE=0
SILENT=0
PREFS=~/.sshTools.cfg
SYSPREFS="/etc/ast.cfg"
USEPW=""
INPUTLEN=1
CLEARCLIP=0
USERN="`whoami`" ## To mimic default ssh behaviour
PWPATH=~/.pws.cfg
PWNRPATH=~/.pws-nrs.cfg
USESUDO=0
ALWAYSSUDO=0
SSHSCRIPT="SshLoginPasswdInteract.exp"
REMEMBERPW=0
CHANGEPW=0
NOREMPW=0
FIRSTASK=1
USEPWSET=0

## Easy/lazy arg handling
IPSTARTNEXT=0
IPSTOPNEXT=0

for ARG in "$@"
do
  if [ $GETNEXT -eq 1 ];
  then
    GETNEXT=0

    USEPW="$ARG"
    USEPWSET=1
    AUTO=1

    continue
  fi

  if [ $IPSTARTNEXT -eq 1 ];
  then
    IPSTARTNEXT=0

    IPSTART="$ARG"

    continue
  fi

  if [ $IPSTOPNEXT -eq 1 ];
  then
    IPSTOPNEXT=0

    IPSTOP="$ARG"

    continue
  fi

  if [ "$ARG" == "-p" ];
  then
    GETNEXT=1
  elif [ "$ARG" == "-v" ] || [ "$ARG" == "--verbose" ];
  then
    VERBOSE=1
  elif [ "$ARG" == "-S" ] || [ "$ARG" == "-r" ] || [ "$ARG" == "--save-pw" ];
  then
    REMEMBERPW=1
  elif [ "$ARG" == "-u" ] || [ "$ARG" == "--use-saved-pw" ];
  then
    REMEMBERPW=1
  elif [ "$ARG" == "-e" ] || [ "$ARG" == "--change-pw" ];
  then
    CHANGEPW=1
  elif [ "$ARG" == "-n" ] || [ "$ARG" == "--no-saved-pw" ];
  then
    NOREMPW=1
    REMEMBERPW=0
  elif [ "$ARG" == "--mk-conf" ];
  then
    mkConf
  elif [ "$ARG" == "--mk-pwf" ];
  then
    mkPw
  elif [ "$ARG" == "-i" -o "$ARG" == "--start-ip" ];
  then
    IPSTARTNEXT=1
  elif [ "$ARG" == "-t" -o "$ARG" == "--stop-ip" ];
  then
    IPSTOPNEXT=1
  elif [ "$ARG" == "-h" ] || [ "$ARG" == "--help" ];
  then
    echo "SSH Wrapper - Arxx SSH Tools <https://arxx.se>."
    echo
    echo "Syntax: $0 [ OPTION ] .. IP_OR_HOSTNAME [ COMMAND ]"
    echo
    echo "Examples:"
    echo "  $ $0 root@172.16.1.1 'df -h'"
    echo "  $ $0 root@172.16.1.1 -p 3"
    echo "  $ $0 -i test@10.10.0.1 -t 15"
    echo
    echo "You can also specify which pw to use with: -p PASSWORD_NUMBER."
    echo
    echo "Commandline options:"
    echo "  --mk-conf          - Make sample Arxx SSH Tools config file."
    echo "  --mk-pwf           - Make sample password file."
    echo "  -v, --verbose      - Verbose mode."
    echo "  -p PWNR            - Connect with password number PWNR."
    echo "  -i, --start-ip     - IP range start (for login)."
    echo "  -t, --stop-ip      - IP range stop (for login)."
    echo "  -r, -S, --save-pw  - Save password this time only (disregard config setting)."
    echo "  -u, --use-saved-pw - Use a saved password this time only (disregard config setting)."
    echo "  -n, --no-saved-pw  - Do not use any saved passwords and do not save this time only."
    echo "  -e, --change-pw    - Change saved password for a saved host this time only."

    exit 0
  else
    C=$((C+1))
    ARGS[$C]="$ARG"
  fi
done

if [ "$REMEMBERPW"  == "1" ];
then
  REMPWSET=1
else
  REMPWSET=0
fi

if [ -n "$IPSTART" -a -z "$IPSTOP" ] || [ -z "$IPSTART" -a -n "$IPSTOP" ];
then
  printEE "Need both start ip and stop ip!"
fi

if [ -z "${ARGS[1]}" ] && [ -z "$IPSTOP" -o -z "$IPSTART" ];
then
  echo "What? Use -h!"
  exit 1
fi

if [ -r "$SYSPREFS" ];
then
  printIC "Using system preferences set in: $SYSPREFS."

  . "$SYSPREFS"
fi

if [ -r "$PREFS" ];
then
  printIC "Using preferences set in: $PREFS."

  . "$PREFS"
else
  printIC "Config file not found ($PREFS) so using defaults or system config instead."
fi

if [ ! -r "$PWPATH" ];
then
  printEE "Cannot read pw file: $PWPATH."
else
  . "$PWPATH"
fi

## Set this here.
ORIGUSER="$USERN" ## Dup

if [ -z "$IPSTART" ];
then
  STARTIP="${ARGS[1]}"
else
  STARTIP="$IPSTART"
fi

if [ -z "`echo "$STARTIP" | grep -soE '\@'`" ];
then
  START="$USERN@$STARTIP"
else
  START="$STARTIP"
fi

if [ $NOREMPW -eq 1 ] || [ $CHANGEPW -eq 1 ];
then
  REMEMBERPW=0
fi

if [ "$REMPWSET"  == "1" ];
then
  REMEMBERPW=1
fi

COMMAND="${ARGS[2]}"
INITIAL="`basename \`echo "$START" | sed 's/\./\//g'\``"
HOSTSEQ="`echo "$START" | sed 's/\./ /g' | awk '{ print $1 "." $2 "." $3 "." }'`"
RETVAL=1
STOP=0

if [ -z "$COMMAND" ];
then
  if [ -z "$IPSTART" -a -z "$IPSTOP" ];
  then
    while [ $RETVAL -ne 0 ] && [ $STOP -eq 0 ];
    do
      ## Get hostname
      HOSTNAMEORIP="`echo "$START" | sed -n 's/.*@//p'`"

      checkHost "$HOSTNAMEORIP"

      askPass

      dirCheck "$HOSTNAMEORIP"

      if [ "$USESUDO" == "1" ];
      then
        SSHSCRIPT="SshLoginPasswdInteractSudo.exp"
      else
        SSHSCRIPT="SshLoginPasswdInteract.exp"
      fi

      if [ -z "`echo "$STARTIP" | grep -soE '\@'`" ];
      then
        START="$USERN@$STARTIP"
      else
        START="$STARTIP"
      fi

      $SCPATH/$SSHSCRIPT "$START" "$SSHPASSWD" 0 "$NEWDIR"

      RETVAL=$?

      if [ $AUTO -eq 1 ];
      then
        ## Stop after first attempt
        STOP=1
      fi
    done
  else
    NUMHOSTS="$IPSTOP"
    for i in `seq $INITIAL $NUMHOSTS`
    do
      HOST="$HOSTSEQ$i"
      RETVAL=1
      STOP=0

      while [ $RETVAL -ne 0 ] && [ $STOP -eq 0 ];
      do
        checkHost "$HOST"

        askPass

        dirCheck "$HOST"

        if [ "$USESUDO" == "1" ];
        then
          SSHSCRIPT="SshLoginPasswdInteractSudo.exp"
        else
          SSHSCRIPT="SshLoginPasswdInteract.exp"
        fi

        if [ -z "`echo "$STARTIP" | grep -soE '\@'`" ];
        then
          START="$USERN@$STARTIP"
        else
          START="$STARTIP"
        fi

        $SCPATH/$SSHSCRIPT "$HOST" "$SSHPASSWD" 0 "$NEWDIR"

        RETVAL=$?

        if [ $AUTO -eq 1 ];
        then
          ## Stop after first attempt
          STOP=1
        fi
      done
    done
  fi
else
  while [ $RETVAL -ne 0 ] && [ $STOP -eq 0 ];
  do
    ## Get hostname
    HOSTNAMEORIP="`echo "$START" | sed -n 's/.*@//p'`"

    checkHost "$HOSTNAMEORIP"

    askPass

    dirCheck "$HOSTNAMEORIP"

    if [ "$USESUDO" == "1" ];
    then
      SSHSCRIPT="SshCmdLoginPasswdSudo.exp"
    else
      SSHSCRIPT="SshCmdLoginPasswd.exp"
    fi

    if [ -z "`echo "$STARTIP" | grep -soE '\@'`" ];
    then
      START="$USERN@$STARTIP"
    else
      START="$STARTIP"
    fi

    $SCPATH/$SSHSCRIPT "$START" "$SSHPASSWD" "$COMMAND" 0

    RETVAL=$?

    if [ $AUTO -eq 1 ];
    then
      ## Stop after first attempt
      STOP=1
    fi
  done
fi
echo "------------------------------------------------------------"
echo "## I: Done!"

if [ ${CLEARCLIP} -eq 1 ];
then
  echo -n "CLEARED" | xclip
fi
