#!/usr/bin/env python

## Part of Arxx SSH Tools. Published at: <https://arxx.se>.
## Written by: Johan Hyllstedt. You can reach me at: <code@arxx.se>.
## See README for more information.

## Copyright 2010, 2011, 2012 Johan Hyllstedt.

## This file is part of Arxx SSH Tools.

## Arxx SSH Tools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## Arxx SSH Tools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with Arxx SSH Tools.  If not, see <http://www.gnu.org/licenses/>.

from subprocess import Popen
import sys
import gtk
import os
import signal
import random
import threading, thread
import time
import ConfigParser
import gobject
import re

gtk.gdk.threads_init()

## Some astgui constants
ON = 1
OFF = 0

INFO = 0
WARN = 1
ERR = 2
NOUPDATE = 213

SAFE = True
UNSAFE = False

RUNNING = 1
FAILED = 2
EXITOK = 3

## Working List constants....
PREV = 1
NEXT = 2

GET = 44
CHK = 46
ADD = 202

FWR = 484
BWR = 482
VAL = 486

def errorDisp(text):
  ## Prints to stderr
  print >>sys.stderr, text

class cmdThread(threading.Thread):
  def run(self):
    mainGui.updateOutput(OFF)
    oldmtime = 0
    mainGui.prevOutfile = mainGui.outfile ## For prev function to work
    mainGui.outfile = mainGui.returnTmpFile()
    status = None

    try:
      mainGui.infoDisp(ON, INFO, "Job started OK.")
      mainGui.updateJobStatus(RUNNING)
      cmdline = str(mainGui.cmdout + " 1>>" + mainGui.outfile + " 2>>" + mainGui.outfile)

      mainGui.sp = Popen(cmdline, shell = True, preexec_fn = os.setsid)
      print "I: Full commandline: %s." % cmdline

      while status == None:
        print "I: Waiting for job to exit..."
        time.sleep(1)

        mtime = os.path.getmtime(mainGui.outfile)

        if mtime > oldmtime:
          print "I: Buffer updated."
          f = open(mainGui.outfile, 'r')
          buff = f.read()
          f.close()
          mainGui.updateOutput(ON, buff)
          oldmtime = mtime

        status = mainGui.sp.poll()

      print "I: Job has exited."

      retcode = mainGui.sp.returncode

      if retcode == 0:
        mainGui.infoDisp(ON, INFO, "Job exited successfully.")
        mainGui.updateJobStatus(EXITOK)
      else:
        mainGui.infoDisp(ON, ERR, "Job exited with errors (or was killed).")
        errorDisp("W: Job exited with errors (or was killed).")
        mainGui.updateJobStatus(FAILED)
    except OSError, e:
      mainGui.infoDisp(ON, ERR, "Execution failed:\n" + str(e))
      errorDisp("W: Execution failed: " + str(e))
      mainGui.updateJobStatus(FAILED)
    except IOError:
      pass

class WorkingList:
  ## As in; a *working* list! :-)
  ## Sort of cstyle "Linked list", sort of...
  ## Always keeps 10 items in list (not at first but as it grows)
  ## Designed to work with INT and STR
  ## Keep internal pointer at current obj, goes, next, prev, first, last
  ## Does add only
  ## Return bool or str depending on method

  def __init__(self):
    self.ptr = -1
    self.count = -1
    self.lType = [ ]
    self.lText = [ ]

  def first(self):
    ## Sets ptr to end (oldest)
    go = True

    while go:
      if self.chk(PREV):
        self.setPtr(BWR)
      else:
        go = False

  def setPtr(self, direction, val = None):
    if direction == FWR:
      self.ptr += 1
    elif direction == BWR:
      self.ptr -= 1
    elif direction == VAL:
      if val != None:
        self.ptr = val

  def last(self):
    ## Sets ptr to begining (newest)
    go = True

    while go:
      if self.chk(NEXT):
        self.setPtr(FWR)
      else:
        go = False

  def chk(self, direction):
    ## Check if possible to go forward or backwards.
    ## Return True or False depending on above.
    if len(self.lType) == 0 or len(self.lText) == 0:
      return False

    if direction == PREV:
      prev = self.getPtr() - 1

      if prev < 0:
        return False

      try:
        tType = self.lType[prev]
        tText = self.lText[prev]
      except:
        return False
      else:
        return True
    elif direction == NEXT:
      try:
        nxt = self.getPtr() + 1
        tType = self.lType[nxt]
        tText = self.lText[nxt]
      except:
        return False
      else:
        return True

  def get(self, direction):
    ## Gets value in direction
    ## Generally you WANT to do a CHK first to get correct results.
    ## Return None if not possible.

    if direction == PREV:
      if self.chk(PREV):
        prev = self.getPtr() - 1
        tType = self.lType[prev]
        tText = self.lText[prev]

        return tType, tText
    elif direction == NEXT:
      if self.chk(NEXT):
        nxt = self.getPtr() + 1
        tType = self.lType[nxt]
        tText = self.lText[nxt]

        return tType, tText

  def next(self):
    self.setPtr(FWR)

  def prev(self):
    self.setPtr(BWR)

  def add(self, dType, dText):
    self.count += 1
    self.last()
    self.setPtr(FWR)
    self.lType.append(dType)
    self.lText.append(dText)

    if self.count >= 10:
      self.setPtr(BWR)
      self.lType.pop(0)
      self.lText.pop(0)

  def getPtr(self):
    ## -1 = Empty
    ## 0 - 9 = Entries
    return int(self.ptr)

class GUI:
  def __init__(self):
    self.missing = False
    self.cmdout = None
    self.outfile = ""
    self.missingText = ""
    self.sp = None
    self.cfgFolder = str(os.getenv("HOME")) + "/.config/astgui/" ## This will hold individual profiles
    self.globalCfg = "/etc/ast.cfg"
    self.cfgPath = self.cfgFolder + ".default.cfg" ## Default config
    self.lastProfilePath = "" ## Holds path to last used/loaded profile that was OK
    self.autoScroll = False
    self.scrolling = False
    self.prevOutfile = ""
    self.ignoreLoadChange = False
    self.loadText = "Load settings:" ## Load combobox text
    self.deleteText = "Delete settings:" ## Load combobox text
    self.cc = re.compile('[^\x09\x0A\x0D\x20-\x7F]')

    self.listobj = None
    self.firstPop = True

    self.builder = gtk.Builder()
    self.builder.add_from_file(self.getPath() + "/Astgui.glade")
    self.builder.connect_signals(self)

    self.wList = WorkingList()
    self.loadDefaultCfg()

    ## Set default values if unset
    if self.getEditor() == "":
      self.setEditor("gedit")

    if self.getTerminal() == "":
      self.setTerminal("gnome-terminal -e")

    if self.builder.get_object("pathtoscripts_entry").get_text() == "":
      print "I: As the path is currently not set to anything, trying to guess it."
      self.builder.get_object("pathtoscripts_entry").set_text(self.getPath(SAFE))

    self.createCfgDirs()
    self.populateLoadCombo()

  def populateLoadCombo(self):
    print "I: Refreshing load combo box."
    ## This fills the load combo box with the configs saved.

    ## While changing items and such, ignore these
    self.ignoreLoadChange = True

    try:
      files = os.listdir(self.cfgFolder)
    except:
      return ## We cannot populate what is not there

    if self.firstPop == True:
      ## Init combo list
      print "I: Doing first init of load combobox."

      self.listobj = gtk.ListStore(gobject.TYPE_STRING)
      cbox = self.builder.get_object("load_combo")
      cbox.set_model(self.listobj)

      cell = gtk.CellRendererText()
      cbox.pack_start(cell, True)
      cbox.add_attribute(cell, 'text', 0)
      self.firstPop = False

    ## Clear
    self.listobj.clear()

    ## New
    if not self.builder.get_object("deleteprofile_tog").get_active():
      self.listobj.append([self.loadText])
    else:
      self.listobj.append([self.deleteText])

    for File in files:
      if File[:1] != "." and File[-4:] == ".cfg":
        File = File.split(".cfg")[0]
        self.listobj.append([File])
        print "I: Populating with: %s." %File

    self.builder.get_object("load_combo").set_active(0)

    self.ignoreLoadChange = False

  def createCfgDirs(self):
   dotConfig = str(os.getenv("HOME")) + "/.config/"
   dotConfigAstgui = dotConfig + "astgui/"

   if not os.path.exists(dotConfig):
     print "I: Making dir: %s." %dotConfig
     try:
       os.mkdir(dotConfig, 0700)
     except:
      print "W: Cannot make config dir: %s." %dotConfig

   if not os.path.exists(dotConfigAstgui):
     print "I: Making dir: %s." %dotConfigAstgui
     try:
       os.mkdir(dotConfigAstgui, 0700)
     except:
      print "W: Cannot make config dir: %s." %dotConfigAstgui

  def on_filemode_tog_toggled(self, widget = None):
    if self.builder.get_object("filemode_tog").get_active():
      print "I: Filemode toggled."

      print "I: Making GUI layout changes.."
      ## Lock stopip_entry
      self.builder.get_object("stopip_entry").set_sensitive(False)

      ## Change startip_label = File
      self.builder.get_object("startip_label").set_text("File:")

      ## Change tooltip
      self.builder.get_object("startip_entry").set_tooltip_text("Enter the path to the file which contains the list of hosts to load.")

      ## Clear stop ip field
      #self.builder.get_object("stopip_entry").set_text("")
    else:
      print "I: Filemode untoggled."

      print "I: Resetting layout changes."
      ## Lock stopip_entry
      self.builder.get_object("stopip_entry").set_sensitive(True)

      ## Change startip_label = Start IP
      self.builder.get_object("startip_label").set_text("Start IP:")

      ## Reset tooltip = "Start ip, or in hostmode the hostname."
      self.builder.get_object("startip_entry").set_tooltip_text("Start ip or in hostmode the hostname.")

  def on_deleteprofile_tog_toggled(self, widget = None):
    self.populateLoadCombo()

  def on_reloadloadcombo_button_clicked(self, widget = None):
    print "I: Reload load combobox button pressed."
    self.populateLoadCombo()

  def on_setdefprofile_button_clicked(self, widget = None):
    print "I: Set default profile request."

    if self.builder.get_object("defaultprofile_entry").get_text() == "":
      print "I: Unset default cfg requested."
      ## Delete request
      empty = True
    else:
      empty = False

    ## Make dir if needed
    self.createCfgDirs()

    cfgFilePath = self.cfgFolder + self.builder.get_object("defaultprofile_entry").get_text() + ".cfg"

    if not os.path.exists(cfgFilePath) and not empty:
      print "I: There is no profile called: %s!" %cfgFilePath
      self.infoDisp(ON, WARN, "Cannot set default profile, as the name you entered does not point to a valid profile!")
      return

    ## Make self.cfgPath
    cfg = ConfigParser.RawConfigParser()
    cfg.add_section('GUI')

    if not empty:
      cfg.set('GUI', 'defaultprofile', str(self.builder.get_object("defaultprofile_entry").get_text()))
    else:
      cfg.set('GUI', 'defaultprofile', '')

    try:
      with open(self.cfgPath, 'w') as configfile:
        cfg.write(configfile)
    except:
      self.infoDisp(ON, WARN, "Failed to change default profile in: " + self.cfgPath + ", message: " + str(sys.exc_info()[0]) + ".")
      print "W: Failed to change default profile in: %s, message: %s." % (cfgFile, str(sys.exc_info()[0]))
    else:
      if not empty:
        self.infoDisp(ON, INFO, "Saved default profile succesfully.")
      else:
        self.infoDisp(ON, INFO, "Unset default profile succesfully.")

  def deleteCfg(self, name):
    ## Delete cfg by name.
    cfgFile = self.cfgFolder + name + ".cfg"

    if os.path.exists(cfgFile):
      ## Del
      print "I: Deleting file: %s." %cfgFile

      try:
        os.unlink(cfgFile)
      except:
        print "I: Delete failed!"
        self.infoDisp(ON, WARN, "Delete profile: " + name + " failed.")
      else:
        print "I: Delete ok."
        self.infoDisp(ON, INFO, "Deleted profile: " + name + " ok.")
    else:
      print "I: The file: %s, does not exist?!" %cfgFile

    ## Reload list
    self.populateLoadCombo()

  def on_load_combo_changed(self, widget = None):
    print "I: Load combo box changed."

    if self.ignoreLoadChange == True:
      print "I: Ignoreing this request."
      return

    model = self.builder.get_object("load_combo").get_model()
    index = self.builder.get_object("load_combo").get_active()

    if index:
      if self.builder.get_object("deleteprofile_tog").get_active():
        print "I: Deleting requested cfg."
        self.deleteCfg(model[index][0])
      else:
        print "I: Loading requested cfg."
        self.loadCfg(model[index][0])
    else:
      print "I: Not doing anything as it does not look like it should."

  def returnTmpFile(self):
    tmpHome = os.getenv("TMPDIR")

    if tmpHome == None or str(tmpHome) == "":
      tmpHome = "/tmp/"
    else:
      tmpHome = str(tmpHome)

    roughdraft = tmpHome + "/astgui_" + str(random.random())

    if os.path.exists(roughdraft):
      roughdraft += str(random.random())

      if os.path.exists(roughdraft):
        errorDisp("E: Cannot seem to create tmp file... Shutting down.")
        self.closeDown()
      else:
        return roughdraft
    else:
      return roughdraft

  def loadDefaultCfg(self):
    ## This will find out if there is a default profile specified, and load it if it is.
    print "I: Load default cfg request."

    if os.path.isfile(self.cfgPath):
      cfg = ConfigParser.RawConfigParser()

      try:
        cfg.read(self.cfgPath) ## I know it might not be readable, but if a user has a cfg file at this path, which is not readble by himself there is a lot of other problems a foot...........
      except:
        ## Looks like this function does not raise any errors even though the files is not readble... well well
        self.infoDisp(ON, WARN, "Cannot read default cfg file: " + self.cfgPath + " even though it exists.")
        print "W: Cannot read default cfg file: %s even though it exists." % self.cfgPath
        return

      try:
        defaultprofile = cfg.get('GUI', 'defaultprofile')
      except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
        self.infoDisp(ON, WARN, "Load problems with the default profile. Probably corrupt cfg file. Try resetting the default profile..")
        print "W: Load problems with the default profile. Probably corrupt cfg file. Try resetting the default profile.."
      else:
        ## Set the value of field
        self.builder.get_object("defaultprofile_entry").set_text(defaultprofile)

        ## Attempt to load it
        if defaultprofile != "":
          self.loadCfg(defaultprofile)
        else:
          print "I: Default profile not set, not loading!"
    else:
      print "I: Cannot find default cfg file, default profile set?"

    print "I: Load default profile procedure done."

  def loadCfg(self, name):
    ## This now tries to load a cfg by name.

    ## Cfg should be saved in ~/.config/astgui/*.cfg
    print "I: Loading settings in profile: %s." %name

    def verifyBoolSetActive(self, section, entry):
      if cfg.getboolean(section, entry) and not self.builder.get_object(entry).get_active():
        self.builder.get_object(entry).clicked()
      elif not cfg.getboolean(section, entry) and self.builder.get_object(entry).get_active():
        self.builder.get_object(entry).clicked()

    def verifyStrSetTxt(self, section, entry):
      self.builder.get_object(entry).set_text(cfg.get(section, entry))

    cfgFile = self.cfgFolder + name + ".cfg"

    if os.path.isfile(cfgFile):
      cfg = ConfigParser.RawConfigParser()

      try:
        cfg.read(cfgFile) ## I know it might not be readable, but if a user has a cfg file at this path, which is not readble by himself there is a lot of other problems a foot...........
      except:
        ## Looks like this function does not raise any errors even though the files is not readble... well well
        self.infoDisp(ON, WARN, "Cannot read cfg file: " + cfgFile + " even though it exists.")
        print "W: Cannot read cfg file: %s even though it exists." % cfgFile
        return

      try:
        ## We must now set all values accordingly.

        ## Width / height
        self.builder.get_object("mainWindow").resize(cfg.getint('GUI', 'w'), cfg.getint('GUI', 'h'))
        ## Active tab
        self.builder.get_object("notebook1").set_current_page(cfg.getint('GUI', 'tab'))
        ## Push/pull
        if cfg.getboolean('GUI', 'push'):
          self.builder.get_object("pushradio").clicked()
        elif cfg.getboolean('GUI', 'pull'):
          self.builder.get_object("pullradio").clicked()

        if cfg.getboolean('GUI', 'fullscreen_tog'):
          self.builder.get_object("mainWindow").maximize()
          verifyBoolSetActive(self, "GUI", "fullscreen_tog")

        ## window pos
        if cfg.getboolean('GUI', 'winpos_tog'):
          self.builder.get_object("mainWindow").move(cfg.getint('GUI', 'x'), cfg.getint('GUI', 'y'))
          verifyBoolSetActive(self, "GUI", "winpos_tog")

        ## scroll bottom
        if cfg.getboolean('GUI', 'scroll_tog'):
          self.autoScroll = True
          verifyBoolSetActive(self, "GUI", "scroll_tog")

        ## And then rest of SSH settings
        verifyBoolSetActive(self, "SSH", "ss_tog")
        verifyBoolSetActive(self, "SSH", "silent_tog")
        verifyBoolSetActive(self, "SSH", "verbose_tog")
        verifyBoolSetActive(self, "SSH", "hostmode_tog")
        verifyBoolSetActive(self, "SSH", "pad_tog")
        verifyBoolSetActive(self, "SSH", "sshkeys_tog")
        verifyBoolSetActive(self, "SSH", "criterr_tog")
        verifyBoolSetActive(self, "SSH", "exiterr_tog")
        verifyBoolSetActive(self, "SSH", "scpmkdirs_tog")
        verifyBoolSetActive(self, "SSH", "filemode_tog")
        verifyBoolSetActive(self, "SSH", "sudo_tog")

        verifyStrSetTxt(self, "SSH", "starthost_entry")
        verifyStrSetTxt(self, "SSH", "pad_entry")
        verifyStrSetTxt(self, "SSH", "startip_entry")
        verifyStrSetTxt(self, "SSH", "stopip_entry")
        verifyStrSetTxt(self, "SSH", "username_entry")
        verifyStrSetTxt(self, "SSH", "password_entry")
        verifyStrSetTxt(self, "SSH", "file_entry")
        verifyStrSetTxt(self, "SSH", "dest_entry")
        verifyStrSetTxt(self, "SSH", "sshopts_entry")
        verifyStrSetTxt(self, "SSH", "command_entry")

        verifyStrSetTxt(self, "GUI", "pathtoscripts_entry")
        verifyStrSetTxt(self, "GUI", "editor_entry")
        verifyStrSetTxt(self, "GUI", "terminal_entry")
      except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
        self.infoDisp(ON, WARN, "Load problems with: " + name + ". Could be: one or more option(s) does not exist in cfg file. Or perhaps it could not load the file. Try resaving.")
        print "W: Problem loading: %s. One or more option(s) does not exist in cfg file... could be bad, or just nothing. Try resaving. Also could be because it could not load the file." %name
      else:
        ## If we could load this ok, keep this path as last.
        self.lastProfilePath = cfgFile ## Keep this path
        self.infoDisp(ON, INFO, "Loading done succesfully of: " + name + ".")
    else:
      print "I: Cannot find cfg file: %s, settings saved?" %cfgFile
      self.infoDisp(ON, WARN, "Cannot find cfg file: " + name + ". Try checking directory permissions and such.")

    print "I: Load procedure done."

  def getEditor(self):
    return str(self.builder.get_object("editor_entry").get_text())

  def getTerminal(self):
    return str(self.builder.get_object("terminal_entry").get_text())

  def setEditor(self, text):
    self.builder.get_object("editor_entry").set_text(text)

  def setTerminal(self, text):
    self.builder.get_object("terminal_entry").set_text(text)

  def getPath(self, safe = False):
    ## Return (str) path to where Astgui.glade are atleast, hopefully the ssh tools too
    ## Does a bit of guess work too....

    ## Check GUI Path
    try:
      guipath = str(self.builder.get_object("pathtoscripts_entry").get_text())
    except:
      if self.lastProfilePath != "":
        if os.path.isfile(self.lastProfilePath):
          try:
            cfg = ConfigParser.RawConfigParser()
            cfg.read(self.lastProfilePath)
            guipath = str(cfg.get('GUI', 'pathtoscripts_entry'))
          except:
            pass
          else:
            if not os.path.exists(guipath + "/Astgui.glade"):
              errorDisp("W: Cannot find Astgui.glade at this path (from cfg file): '" + guipath + "'.")
            else:
              return str(guipath)
    else:
      if not os.path.exists(guipath + "/Astgui.glade"):
        errorDisp("W: Cannot find Astgui.glade at this path (entered in gui): '" + guipath + "'.")
      else:
        return str(guipath)

    ## Try to find it
    candidatePath = os.path.abspath(os.path.dirname(__file__))
    if not os.path.exists(candidatePath + "/Astgui.glade"):
      errorDisp("W: Cannot find Astgui.glade at this path (gussed from the location of this script): '" + candidatePath + "'.")
    else:
      return str(candidatePath)

    ## Try with global config file
    if os.path.exists(self.globalCfg):
      try:
        with open(self.globalCfg, 'r') as globCfgFile:
          globCfg = globCfgFile.read()
      except:
        pass
      else:
        if globCfg != None and globCfg != "":
          search = re.search("SCPATH=.+", globCfg)

          if search != None:
            match = str(search.group(0))
            guipath = match.split("=")[1]
            guipath = os.path.expanduser(guipath)
            if os.path.exists(guipath):
              return str(guipath)
            else:
              errorDisp("W: Cannot find astgui here: '" + guipath + "', as entered in global config file.")

    ## Try with first param
    try:
      inputparam = sys.argv[1]
    except:
      pass
    else:
      if not os.path.exists(inputparam + "/Astgui.glade"):
        errorDisp("W: Cannot find Astgui.glade at this path (entered as first parameter): '" + inputparam + "'.")
      else:
        return str(inputparam)

    if not safe:
      errorDisp("E: Tried my best to find required file Astgui.glade. Giving up.\nI: Try entering the path to Astgui.glade and Arxx SSH Tools as the first argument.")
      sys.exit(1)
    else:
      return str("")

  def saveCfg(self, name):
    ## Now to use with a name.

    ## Make dir if needed
    self.createCfgDirs()

    cfgFile = self.cfgFolder + name + ".cfg"

    if os.path.exists(cfgFile):
      print "W: The file: %s already exists, trying to find next available number." % cfgFile

      incr = 0
      origname = name

      while os.path.exists(cfgFile) and incr <= 99:
        incr += 1
        cfgFile = self.cfgFolder + origname + "-" + str(incr) + ".cfg"
        name = origname + "-" + str(incr)

      if incr >= 100:
        print "E: Failed to find a new name for profile, tried up to 99 new names!"
        self.closeDown()

      print "I: New profile name is: %s." % cfgFile


    print "I: Saveing settings to: %s (%s)." % (name, cfgFile)

    cfg = ConfigParser.RawConfigParser()
    cfg.add_section('GUI')
    cfg.set('GUI', 'tab', int(self.builder.get_object("notebook1").get_current_page()))
    cfg.set('GUI', 'w', int(self.builder.get_object("mainWindow").get_size()[0]))
    cfg.set('GUI', 'h', int(self.builder.get_object("mainWindow").get_size()[1]))
    cfg.set('GUI', 'push', bool(self.builder.get_object("pushradio").get_active()))
    cfg.set('GUI', 'pull', bool(self.builder.get_object("pullradio").get_active()))
    cfg.set('GUI', 'fullscreen_tog', bool(self.builder.get_object("fullscreen_tog").get_active()))
    cfg.set('GUI', 'winpos_tog', bool(self.builder.get_object("winpos_tog").get_active()))
    cfg.set('GUI', 'x', int(self.builder.get_object("mainWindow").get_position()[0]))
    cfg.set('GUI', 'y', int(self.builder.get_object("mainWindow").get_position()[1]))
    cfg.set('GUI', 'scroll_tog', bool(self.builder.get_object("scroll_tog").get_active()))
    cfg.set('GUI', 'pathtoscripts_entry', str(self.builder.get_object("pathtoscripts_entry").get_text()))
    cfg.set('GUI', 'editor_entry', str(self.builder.get_object("editor_entry").get_text()))
    cfg.set('GUI', 'terminal_entry', str(self.builder.get_object("terminal_entry").get_text()))

    cfg.add_section('SSH')
    cfg.set('SSH', 'ss_tog', bool(self.builder.get_object("ss_tog").get_active()))
    cfg.set('SSH', 'silent_tog', bool(self.builder.get_object("silent_tog").get_active()))
    cfg.set('SSH', 'verbose_tog', bool(self.builder.get_object("verbose_tog").get_active()))
    cfg.set('SSH', 'hostmode_tog', bool(self.builder.get_object("hostmode_tog").get_active()))
    cfg.set('SSH', 'pad_tog', bool(self.builder.get_object("pad_tog").get_active()))
    cfg.set('SSH', 'sshkeys_tog', bool(self.builder.get_object("sshkeys_tog").get_active()))
    cfg.set('SSH', 'criterr_tog', bool(self.builder.get_object("criterr_tog").get_active()))
    cfg.set('SSH', 'exiterr_tog', bool(self.builder.get_object("exiterr_tog").get_active()))
    cfg.set('SSH', 'scpmkdirs_tog', bool(self.builder.get_object("scpmkdirs_tog").get_active()))
    cfg.set('SSH', 'filemode_tog', bool(self.builder.get_object("filemode_tog").get_active()))
    cfg.set('SSH', 'sudo_tog', bool(self.builder.get_object("filemode_tog").get_active()))

    cfg.set('SSH', 'starthost_entry', str(self.builder.get_object("starthost_entry").get_text()))
    cfg.set('SSH', 'pad_entry', str(self.builder.get_object("pad_entry").get_text()))
    cfg.set('SSH', 'startip_entry', str(self.builder.get_object("startip_entry").get_text()))
    cfg.set('SSH', 'stopip_entry', str(self.builder.get_object("stopip_entry").get_text()))
    cfg.set('SSH', 'username_entry', str(self.builder.get_object("username_entry").get_text()))
    cfg.set('SSH', 'password_entry', str(self.builder.get_object("password_entry").get_text()))

    cfg.set('SSH', 'file_entry', str(self.builder.get_object("file_entry").get_text()))
    cfg.set('SSH', 'dest_entry', str(self.builder.get_object("dest_entry").get_text()))
    cfg.set('SSH', 'sshopts_entry', str(self.builder.get_object("sshopts_entry").get_text()))

    cfg.set('SSH', 'command_entry', str(self.builder.get_object("command_entry").get_text()))

    try:
      with open(cfgFile, 'w') as configfile:
        cfg.write(configfile)
    except:
      self.infoDisp(ON, WARN, "Failed to save settings to: " + cfgFile + ", message: " + str(sys.exc_info()[0]) + ".")
      print "W: Failed to save settings to: %s, message: %s." % (cfgFile, str(sys.exc_info()[0]))
    else:
      self.infoDisp(ON, INFO, "Saved: " + name + " succesfully.")

    print "I: Save procedure done."
    self.populateLoadCombo()

  def closeDown(self):
    print "I: Closing down."

    self.cancelJob()
    gtk.main_quit()
    sys.exit(0);

  def cancelJob(self):
    print "I: Killing any active job."

    if self.jobRunning():
      os.killpg(self.sp.pid, signal.SIGKILL)
      print "I: Cancel on pid: %s." % str(self.sp.pid)
    else:
      print "I: No job running."

  def on_mainWindow_destroy(self, widget = None):
    self.closeDown()

  def on_save_button_clicked(self, widget = None):
    if self.builder.get_object("newprofile_entry").get_text() == "":
      newname = "profile"
    else:
      newname = str(self.builder.get_object("newprofile_entry").get_text())

    self.saveCfg(newname)

  def on_prevopen_button_clicked(self, widget = None):
    if self.prevOutfile != "":
      if os.path.exists(self.prevOutfile):
        os.system(self.getEditor() + " " + self.prevOutfile + "&")
    else:
      print "I: There is no previous output file."

  def on_load_button_clicked(self, widget = None):
    self.loadCfg()

  def on_cancel_button_clicked(self, widget = None):
    self.cancelJob()

  def jobRunning(self):
    try:
      if os.path.exists("/proc/" + str(self.sp.pid)):
        return True
      else:
        print "I: PID %s (/proc/%s) does not exist." % (str(self.sp.pid), str(self.sp.pid))
        return False
    except:
      return False

  def on_open_button_clicked(self, widget, data = None):
    if os.path.exists(self.outfile):
      os.system(self.getEditor() + " " + self.outfile + "&")

  def on_disptext_scroll_event(self, widget, data = None):
    if data.direction == gtk.gdk.SCROLL_UP:
      self.alternateDisp(PREV)
    elif data.direction == gtk.gdk.SCROLL_DOWN:
      self.alternateDisp(NEXT)

  def on_scroll_tog_clicked(self, widget, data = None):
    if widget.get_active():
      print "I: Setting autoscroll ON."
      self.autoScroll = True
    else:
      self.autoScroll = False
      print "I: Setting autoscroll OFF."

  def alternateDisp(self, direction):
    if direction == PREV:
      if not self.wList.chk(PREV):
        self.builder.get_object("disptext").set_tooltip_text("Scroll to switch between previous messages. [At last]")
        print "I: No previous message."
      else:
        print "I: Switching to previous message."
        self.builder.get_object("disptext").set_tooltip_text("Scroll to switch between previous messages.")
        self.infoDisp(ON, self.wList.get(PREV)[0], self.wList.get(PREV)[1], NOUPDATE)
        self.wList.prev()
    elif direction == NEXT:
      if not self.wList.chk(NEXT):
        print "I: No next message."
        self.builder.get_object("disptext").set_tooltip_text("Scroll to switch between previous messages. [At newest]")
      else:
        print "I: Switching to next message."
        self.builder.get_object("disptext").set_tooltip_text("Scroll to switch between previous messages.")
        self.infoDisp(ON, self.wList.get(NEXT)[0], self.wList.get(NEXT)[1], NOUPDATE)
        self.wList.next()

  def infoDisp(self, onoff, disptype = INFO, text = "Something to display?", update = 1):
    if update != NOUPDATE:
      self.wList.add(disptype, text)

    if onoff == ON:
      if disptype == INFO:
        self.builder.get_object("dispimg").set_from_stock(gtk.STOCK_YES, gtk.ICON_SIZE_DND)
      elif disptype == WARN:
        self.builder.get_object("dispimg").set_from_stock(gtk.STOCK_DIALOG_WARNING, gtk.ICON_SIZE_DND)
      else:
        self.builder.get_object("dispimg").set_from_stock(gtk.STOCK_DIALOG_ERROR, gtk.ICON_SIZE_DND)

      self.builder.get_object("disptext").set_text(text)
      self.builder.get_object("disptext").show()
      self.builder.get_object("dispimg").show()
    elif onoff == OFF:
      self.builder.get_object("disptext").hide()
      self.builder.get_object("dispimg").hide()
      self.missing = False

  def updateJobStatus(self, status):
    if status == RUNNING:
      self.builder.get_object("jobstatus").set_from_stock(gtk.STOCK_EXECUTE, gtk.ICON_SIZE_BUTTON)
      self.builder.get_object("jobstatus").set_tooltip_text("A job is running.")
    elif status == FAILED:
      self.builder.get_object("jobstatus").set_from_stock(gtk.STOCK_DIALOG_ERROR, gtk.ICON_SIZE_BUTTON)
      self.builder.get_object("jobstatus").set_tooltip_text("Last job exited with errors.")
    elif status == EXITOK:
      self.builder.get_object("jobstatus").set_from_stock(gtk.STOCK_YES, gtk.ICON_SIZE_BUTTON)
      self.builder.get_object("jobstatus").set_tooltip_text("Last job exited OK.")

  def updateOutput(self, onoff, text = ""):
    ## Remove annoying control characters
    text = self.cc.sub('', text)
    text = text.replace(']0;', '') ## Last bit of crap...
    text = text.replace('[H[2J', '') ## Even more crap.

    if onoff == ON:
      if not self.isScrollBottom():
        self.scrolling = True
      else:
        self.scrolling = False

      self.builder.get_object("output").set_text(text)

      if self.autoScroll and not self.scrolling:
        print "I: Autoscroll when ever GTK gets around to it (sarcasm)."
        gobject.idle_add(mainGui.autoScrollNow)
    else:
      self.builder.get_object("output").set_text("")

  def autoScrollNow(self):
    sadj = self.builder.get_object("adjustment1")
    stext = self.builder.get_object("scrolledwindow1")

    sbar = stext.get_vscrollbar()
    sbar.set_value(sadj.upper - sadj.page_size)

  def isScrollBottom(self):
    sadj = self.builder.get_object("adjustment1")
    stext = self.builder.get_object("scrolledwindow1")

    sbar = stext.get_vscrollbar()
    value = sbar.get_value()

    if value == sadj.upper - sadj.page_size:
      return True
    else:
      return False

  def checkMissingEntry(self, item, flag):
    if str(self.builder.get_object(item + "_entry").get_text()) == "":
      self.addMissing("Missing " + item + ".")
    else:
      self.cmdout += " " + flag + " " + str(self.builder.get_object(item + "_entry").get_text())

  def on_scprun_button_clicked(self, widget, data = None):
    print "I: SCP run request."
    if not self.jobRunning():
      self.cmdout = self.getPath(SAFE) + "/astscp"

      self.resetMissing()

      if not os.path.exists(self.cmdout):
        self.addMissing("Path to scripts is incorrect.")

      self.verifyBasics()

      if self.missing:
        self.infoDisp(ON, ERR, self.missingText)
        print "W: Missing stuff so not running."
        return

      ## Verify rest
      self.checkMissingEntry("file", "-f")
      self.checkMissingEntry("dest", "-d")

      if str(self.builder.get_object("sshopts_entry").get_text()) != "":
        self.cmdout += " -o " + str(self.builder.get_object("sshopts_entry").get_text())

      if self.builder.get_object("scpmkdirs_tog").get_active():
        self.cmdout += " -M"

      if self.builder.get_object("pushradio").get_active() and not self.builder.get_object("pullradio").get_active():
        self.cmdout += " -m h"
      elif not self.builder.get_object("pushradio").get_active() and self.builder.get_object("pullradio").get_active():
        self.cmdout += " -m l"
      else:
        errorDisp("Wtf push & pull cannot be enabled at the same time???")
        self.closeDown()

      if not self.missing:
        print "I: Running cmd in new thread."
        cmdThread().start()
      else:
        self.infoDisp(ON, ERR, self.missingText)
        print "W: Missing stuff so not running."
    else:
      mainGui.infoDisp(ON, WARN, "A job is already running!")

  def on_cmdrun_button_clicked(self, widget, data = None):
    print "I: CMD run request."
    if not self.jobRunning():
      self.cmdout = self.getPath(SAFE) + "/astcmd"

      self.resetMissing()

      if not os.path.exists(self.cmdout):
        self.addMissing("Path to scripts is incorrect.")

      self.verifyBasics()

      if self.missing:
        self.infoDisp(ON, ERR, self.missingText)
        print "W: Missing stuff so not running."
        return

      ## Verify rest
      if str(self.builder.get_object("command_entry").get_text()) == "":
        self.addMissing("Missing command.")
      else:
        self.cmdout += " -c '" + str(self.builder.get_object("command_entry").get_text()) + "'"

      if not self.missing:
        print "I: Running cmd in new thread."
        cmdThread().start()
      else:
        self.infoDisp(ON, ERR, self.missingText)
        print "W: Missing stuff so not running."
    else:
      mainGui.infoDisp(ON, WARN, "A job is already running!")

  def on_loginrun_button_clicked(self, widget, data = None):
    print "I: Login run request."
    if not self.jobRunning():
      self.cmdout = self.getTerminal() + " '" + self.getPath(SAFE) + "/astlogin"

      self.resetMissing()

      if not os.path.exists(self.getPath(SAFE) + "/astlogin"):
        self.addMissing("Path to scripts is incorrect.")

      self.verifyBasics()

      if self.missing:
        self.infoDisp(ON, ERR, self.missingText)
        print "W: Missing stuff so not running."
      else:
        print "I: Running cmd in new thread."
        self.cmdout += "'"
        cmdThread().start()
    else:
      mainGui.infoDisp(ON, WARN, "A job is already running!")

  def on_loginhelp_button_clicked(self, widget, data = None):
    if not self.jobRunning():
      self.cmdout = self.getPath(SAFE) + "/astlogin"

      if not os.path.exists(self.cmdout):
        self.infoDisp(ON, ERR, "Path to scripts is incorrect.")
      else:
        self.cmdout += " -h"
        cmdThread().start()
    else:
      mainGui.infoDisp(ON, WARN, "A job is already running!")

  def on_scphelp_button_clicked(self, widget, data = None):
    if not self.jobRunning():
      self.cmdout = self.getPath(SAFE) + "/astscp"

      if not os.path.exists(self.cmdout):
        self.infoDisp(ON, ERR, "Path to scripts is incorrect.")
      else:
        self.cmdout += " -h"
        cmdThread().start()
    else:
      mainGui.infoDisp(ON, WARN, "A job is already running!")

  def on_cmdhelp_button_clicked(self, widget, data = None):
    if not self.jobRunning():
      self.cmdout = self.getPath(SAFE) + "/astcmd"

      if not os.path.exists(self.cmdout):
        self.infoDisp(ON, ERR, "Path to scripts is incorrect.")
      else:
        self.cmdout += " -h"
        cmdThread().start()
    else:
      mainGui.infoDisp(ON, WARN, "A job is already running!")

  def addMissing(self, text):
    self.missing = True
    if self.missingText == "":
      self.missingText = text
    else:
      self.missingText += "\n" + text

  def resetMissing(self):
    self.missingText = ""
    self.missing = False

  def verifyBasics(self):
    ## Not mandatory
    if self.builder.get_object("ss_tog").get_active():
      self.cmdout += " -ss"

    if self.builder.get_object("silent_tog").get_active():
      self.cmdout += " -s"

    if self.builder.get_object("verbose_tog").get_active():
      self.cmdout += " -v"

    if self.builder.get_object("hostmode_tog").get_active():
      self.cmdout += " -H"

    if self.builder.get_object("pad_tog").get_active():
      self.cmdout += " -p"

    if self.builder.get_object("sshkeys_tog").get_active():
      self.cmdout += " -k"

    if self.builder.get_object("criterr_tog").get_active():
      self.cmdout += " -C"

    if self.builder.get_object("exiterr_tog").get_active():
      self.cmdout += " -E"

    if self.builder.get_object("sudo_tog").get_active():
      self.cmdout += " -U"

    if self.builder.get_object("filemode_tog").get_active():
      self.cmdout += " -F " + str(self.builder.get_object("startip_entry").get_text())

    if str(self.builder.get_object("username_entry").get_text()) != "":
      username = str(self.builder.get_object("username_entry").get_text()) + "@"
    else:
      username = ""

    if str(self.builder.get_object("starthost_entry").get_text()) != "":
      self.cmdout += " -S " + str(self.builder.get_object("starthost_entry").get_text())

    if str(self.builder.get_object("pad_entry").get_text()) != "":
      self.cmdout += " -n " + str(self.builder.get_object("pad_entry").get_text())

    ## Mandatory
    if str(self.builder.get_object("startip_entry").get_text()) == "":
      if not self.builder.get_object("filemode_tog").get_active():
        self.addMissing("Missing startip.")
      else:
        self.addMissing("Missing file.")
    else:
      if not self.builder.get_object("filemode_tog").get_active():
        self.cmdout += " -i '" + username + str(self.builder.get_object("startip_entry").get_text()) +"'"

    if str(self.builder.get_object("stopip_entry").get_text()) == "" and not self.builder.get_object("filemode_tog").get_active():
      self.addMissing("Missing stopip.")
    else:
      if not self.builder.get_object("filemode_tog").get_active():
        self.cmdout += " -t " + str(self.builder.get_object("stopip_entry").get_text())

    if not self.builder.get_object("sshkeys_tog").get_active() and str(self.builder.get_object("password_entry").get_text()) == "":
      self.addMissing("Missing password (keys not enabled).")
    else:
      if not self.builder.get_object("sshkeys_tog").get_active():
        self.cmdout += " -P " + str(self.builder.get_object("password_entry").get_text())

    return True

if __name__ == "__main__":
  try:
    firstarg = sys.argv[1]
  except:
    pass
  else:
    if firstarg == "-h" or firstarg == "--help" or firstarg == "-help":
      print "This is the GUI frontend for Arxx SSH Tools. (https://arxx.se)"
      print
      print "COMMANDLINE HELP"
      print "  It only takes one argument which is normally not necessary, anyhow that argument is the path to the file called 'Astgui.glade'."
      print
      print "GENERAL"
      print "  Basically what it does is it to invoke Arxx SSH Tools with the arguments you choose, but in a graphical interface instead of the shell."
      print "  A small note; this GUI saves it preferences separately from the commandline tools, so keep that in mind if you get \"weird\" results."

      sys.exit(0)

  mainGui = GUI()
  gtk.main()
