#!/bin/bash

SCPATH=~/sshTools ## Path to SSH Tools. No trailing slash. This can also be set in the configuration file.

## Part of Arxx SSH Tools. Published at: <https://arxx.se>.
## Written by: Johan Hyllstedt. You can reach me at: <code@arxx.se>.
## See README for more information.

## Copyright 2010, 2011, 2012 Johan Hyllstedt.

## This file is part of Arxx SSH Tools.

## Arxx SSH Tools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## Arxx SSH Tools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with Arxx SSH Tools.  If not, see <http://www.gnu.org/licenses/>.

function printEE
{
  ## Print Error & Exit
  ## Prints to stdout & stderr and exits

  echo -e "E: ${1:-"Unknown Error"}" 1>&2

  exit 1
}

function printWC
{
  ## Print Warning & Continues
  ## Prints to stdout & stderr

  if [ $SILENT -eq 0 ];
  then
    if [ -n "$2" ];
    then
      echo -e "${1:-"Unknown Warning"}" 1>&2
    else
      echo -e "W: ${1:-"Unknown Warning"}" 1>&2
    fi
  fi
}

function printIC
{
  ## Print Information & Continues, but only if vebose

  if [ $VERBOSE -eq 1 ] && [ $SILENT -eq 0 ];
  then
    echo -e "I: ${1:-"Unknown Information"}"
  fi
}

function mkConf
{
  if [ -f "$PREFS" ];
  then
    printEE "The file '$PREFS' already exists, I will not overwrite this!"
  fi

  echo "I: Making sample configuration file in: '$PREFS'."

  cat >"$PREFS"<<EOF
## Sample configuration file for Arxx SSH Tools.
## The values set here at the beginning are the default values used by the tools.
## Note that most values here can be overridden at runtime by specifying other values in the input parameters.
## Comments like this are ignored.

## You can also set the path to the tools here. No trailing slash.
SCPATH=~/sshTools

## To always use SILENT mode by default set this to 1.
SILENT=0
SSILENT=0         ## Also set this to 1 for super-silent mode (-ss).

## To always use VERBOSE mode by default set this to 1 (disable SILENT above first).
VERBOSE=0

## To always use pad by default set this to 1.
ENABLEPAD=0

## To use a specific pad length set this to [NR]d, so for 5 it will be "5d".
PAD="2d"

## To always use only SSH keys by default set this to 1.
KEYSONLY=0

## To always use HOSTMODE per default set this to 1.
HOSTMODE=0

## To specify a default password you can do this here, note that KEYSONLY must be 0.
## This naturally also sets password mode as the default.
## Also note that anyone who can read this file can also get the password of course!
SSHPASSWD=""

## To use CRITICALERROR mode by default set this to 1.
CRITICALERROR=0

## To use EXITERROR mode by default set this to 1.
EXITERROR=0

## To always use FILEMODE per default and load hosts from file, set these below.
FILEMODE=0
FILE=""

## To always sudo to root first set this to 1 (does not work with SCP right now).
USESUDO=0

## Paths to SSH binaries. Usually you can leave these as they are.
SCPBIN="/usr/bin/scp"
SSHBIN="/usr/bin/ssh"

## Send these arguments to SCP instead of the default. No starting dash needed.
INPUTSSHOPTS=""
EOF

  exit 0
}

function ipConnect
{
  HOST="$HOSTSEQ$1"

  if [ "$USESUDO" == "1" ];
  then
    CONSCRIPT="SshLoginPasswdInteractSudo.exp"
  else
    CONSCRIPT="SshLoginPasswdInteract.exp"
  fi

  if [ -z "$SSHPASSWD" ];
  then
    printIC "Connecting with SSH normally. With '$SSHOPTS' args."

    $SSHBIN "$HOST" "$SSHOPTS"
  else
    printIC "Connecting with SSH password auth. With '$EXPOPTS' args."

    if [ $SSILENT -eq 0 ];
    then
      $SCPATH/$CONSCRIPT "$HOST" "$SSHPASSWD" "$EXPOPTS"
    else
      printIC "Super silent selected so grepping away password line(s)."

      ## Changed to this because of annoying bash variable behaviour.........
      EXTRAFILE="${TMPDIR:-"/tmp"}/tmpRetVal-$RANDOM-$$"
      touch "$EXTRAFILE" 1>/dev/null 2>/dev/null || printEE "Failed to set tmp retval file: $EXTRAFILE, for use with -ss to detect retval of child proc."

      { $SCPATH/$CONSCRIPT "$HOST" "$SSHPASSWD" "$EXPOPTS" || echo $? > "$EXTRAFILE"; } | grep -vsE "[pP]assword:( )?.?$" | grep -vsE "^The authenticity of host '.*?' can't be established\..$" | grep -vsE "^RSA key fingerprint is .*?\..$" | grep -vsE "^Are you sure you want to continue connecting \(yes/no\)\? (yes)|(no).$"
    fi
  fi

  LASTRETVAL=$?

  if [ $SSILENT -eq 1 ];
  then
    EXTRARET="`cat "$EXTRAFILE" 2>/dev/null || echo "FAILED"`"
    if [ "$EXTRARET" == "FAILED" ];
    then
      printEE "Failed to get retval from file: $EXTRAFILE."
    fi
    rm "$EXTRAFILE"
  fi

  EXTRARET=${EXTRARET:-0}

  ## To fix stupid behaviour with grep and no output
  if [ $SSILENT -eq 1 ] && [ $LASTRETVAL -eq 1 ] && [ $EXTRARET -eq 0 ];
  then
    LASTRETVAL=0
  fi

  if [ $EXTRARET -gt 0 ];
  then
    LASTRETVAL=$EXTRARET
  fi

  if [ $LASTRETVAL -gt 0 ] && [ $CRITICALERROR -eq 1 ];
  then
    printIC "Critical error set and exit status was just above 0. Permanent failure!"

    PERMFAIL=1
  elif [ $LASTRETVAL -gt 0 ] && [ $EXITERROR -eq 1 ];
  then
    printEE "Exit error set and exit status was just above 0. Exiting now."
  fi
}

function hostConnect
{
  if [ "$USESUDO" == "1" ];
  then
    CONSCRIPT="SshLoginPasswdInteractSudo.exp"
  else
    CONSCRIPT="SshLoginPasswdInteract.exp"
  fi

  i="$1"

  if [ $ENABLEPAD -eq 0 ];
  then
    HOST="$START$1"
  else
    STRPAD="`printf "%0$PAD" $1`"
    HOST="$START$STRPAD"
  fi

  if [ -z "$SSHPASSWD" ];
  then
    printIC "Connecting with SSH normally. With '$SSHOPTS' args."

    $SSHBIN "$HOST" "$SSHOPTS"
  else
    printIC "Connecting with SSH password auth. With '$EXPOPTS' args."

    if [ $SSILENT -eq 0 ];
    then
      $SCPATH/$CONSCRIPT "$HOST" "$SSHPASSWD" "$EXPOPTS"
    else
      printIC "Super silent selected so grepping away password line(s)."

      ## Changed to this because of annoying bash variable behaviour.........
      EXTRAFILE="${TMPDIR:-"/tmp"}/tmpRetVal-$RANDOM-$$"
      touch "$EXTRAFILE" 1>/dev/null 2>/dev/null || printEE "Failed to set tmp retval file: $EXTRAFILE, for use with -ss to detect retval of child proc."

      { $SCPATH/$CONSCRIPT "$HOST" "$SSHPASSWD" "$EXPOPTS" || echo $? > "$EXTRAFILE"; } | grep -vsE "[pP]assword:( )?.?$" | grep -vsE "^The authenticity of host '.*?' can't be established\..$" | grep -vsE "^RSA key fingerprint is .*?\..$" | grep -vsE "^Are you sure you want to continue connecting \(yes/no\)\? (yes)|(no).$" 
    fi
  fi

  LASTRETVAL=$?

  if [ $SSILENT -eq 1 ];
  then
    EXTRARET="`cat "$EXTRAFILE" 2>/dev/null || echo "FAILED"`"
    if [ "$EXTRARET" == "FAILED" ];
    then
      printEE "Failed to get retval from file: $EXTRAFILE."
    fi
    rm "$EXTRAFILE"
  fi

  EXTRARET=${EXTRARET:-0}

  ## To fix stupid behaviour with grep and no output
  if [ $SSILENT -eq 1 ] && [ $LASTRETVAL -eq 1 ] && [ $EXTRARET -eq 0 ];
  then
    LASTRETVAL=0
  fi

  if [ $EXTRARET -gt 0 ];
  then
    LASTRETVAL=$EXTRARET
  fi

  if [ $LASTRETVAL -gt 0 ] && [ $CRITICALERROR -eq 1 ];
  then
    printIC "Critical error set and exit status was just above 0. Permanent failure!"

    PERMFAIL=1
  elif [ $LASTRETVAL -gt 0 ] && [ $EXITERROR -eq 1 ];
  then
    printEE "Exit error set and exit status was just above 0. Exiting now."
  fi
}

START=""
NUMHOSTS=""
SSHPASSWD=""
HOSTMODE=0
VERBOSE=0
SILENT=0
SSILENT=0
PAD="2d"
KEYSONLY=0
STARTHOST="1"
ENABLEPAD=0
GO=0
PREFS=~/.sshTools.cfg
SYSPREFS="/etc/ast.cfg"
CRITICALERROR=0
EXITERROR=0
FILE=""
FILEMODE=0
USESUDO=0
SCPBIN="/usr/bin/scp"
SSHBIN="/usr/bin/ssh"

if [ -r "$SYSPREFS" ];
then
  printIC "Using system preferences set in: $SYSPREFS."

  . "$SYSPREFS"
fi

if [ -r "$PREFS" ];
then
  printIC "Using preferences set in: $PREFS."

  . "$PREFS"
else
  printIC "Config file not found ($PREFS) so using defaults or system config instead."
fi

while [ $# -gt 0 ];
do
  case "$1" in
    -h|-help|--help)
      echo "COMMANDLINE SYNTAX"
      echo "  $0 [ OPTIONS ] ... START_IP STOP_IP ..."
      echo "  You can either specify the above parameters directly or by using switches, but you must have"
      echo "  at least a START_IP and STOP_IP, otherwise the script will not run."
      echo
      echo "  Like: $0 root@172.16.1.1 10"
      echo "  Or this works just fine too:"
      echo "        $0 -i root@172.16.1.1 -t 10"
      echo "  If you so choose, it is up to you."
      echo
      echo "COMMANDLINE OPTIONS"
      echo "  -h, -help, --help         - Displays this text."
      echo "  -v, --verbose             - Verbose mode, will output a bit more."
      echo "  -s, -ss, --silent         - Silent mode, will only output errors, -ss for most silent."
      echo "  --mk-conf                 - Generate a sample configuration file in ~/.sshTools.cfg."
      echo "  -H, --host-mode           - Enable hostname mode, see more below."
      echo "  -C, --critical-error      - If any action fails set exit status accordingly always."
      echo "  -E, --exit-error          - Exit immediately if any action fails."
      echo "  -S, --start-host          - Start from X instead of 1 in hostname mode."
      echo "  -p, --pad                 - Enable padding in hostname mode."
      echo "  -n, --nr-pad              - Pad so it is of at least X length."
      echo "  -P, --password            - Enable password mode and use X as password."
      echo "  -i, --start-ip            - Start ip, or in hostmode the hostname."
      echo "  -t, --stop-ip             - The ip to stop at, the last ip address in the list."
      echo "  -k, --keys                - Use SSH keys only."
      echo "  -F, --file                - Use file mode and use the specifed file to load hosts from."
      echo "  -U, --sudo                - First sudo to root using the same password as you login with."
      echo "  --no-sudo                 - Do not sudo to root first."
      echo
      echo "GENERAL"
      echo "  This script is a part of Arxx SSH Tools (https://arxx.se)."
      echo "  And they are intended to simplify management of several hundreds of SSH machines at once and also"
      echo "  with just password authentication. So one does not have to setup SSH keys to be able to manage all"
      echo "  machines. Yes it is not as secure, but that is up to the administrator anyway. Because SSH keys"
      echo "  alone will never stop anyone, so why not be able to automate things with password auth?"
      echo "  Of course it also works fine with just SSH keys."
      echo "  By default the exit value will be the status of the last thing done (see critical-error)."
      echo
      echo "HOSTNAME MODE"
      echo "  This mode is designed to work with several hostnames instead of IP-addresses. It works with the Pad"
      echo "  parameter if you enable it, to increment the hostname to the correct length. So if you enter somehost"
      echo "  as the start host, and 5 as the stop host. Then this script will connect to somehost01 & somehost02 and"
      echo "  so forth until it reaches somehost05. It is also possible to specify a different starting point with the"
      echo "  Start parameter, so per say if you wish to start of at 5 instead that is no problem."
      echo
      echo "FILE MODE"
      echo "  This mode is designed to work with a list of several IP-addresses/hostnames, padding does not work of course."
      echo "  There is no maximum amount of IPs but I would guesstimate that one would not want to run a single command,"
      echo "  with more than 500 IPs. So if you want more than that you can split it up into several smaller runs instead."
      echo "  But theoretically there is no limit. The list it self should be newline or space separated."
      echo "  In the form user@host, unless of course the username of the user that is running the script should be used."
      echo "  You cannot have a space in an individual IP/hostname since that would look like two posts."
      echo
      echo "TROUBLESHOOTING"
      echo "  If you get errors of files not being found, make sure you've either put the tools in ~/sshTools,"
      echo "  or that you've correctly set the SCPATH variable at the top of each file (possible through config file too)."
      echo "  You can also give it a try with these commands if you cant get it running;"
      echo "  chmod u+x $0 && $0 (makes it executable and then runs it)."
      echo "  If it still wont work for you then run it through bash in debug mode like this;"
      echo "  bash -x $0. And it might give you a reason why..."
      echo
      echo "KNOWN BUGS"
      echo "  See HISTORY file."
      echo
      echo "DEFAULT VALUES"
      echo "  |Hostmode) Off. |Pad) 2 digits long."
      echo "  |Password/key auth) Ask. |Pad on/off) off."

      exit
    ;;
    -v|--verbose)
      VERBOSE=1
      SILENT=0
    ;;
    -s|--silent)
      SILENT=1
      VERBOSE=0
    ;;
    -ss)
      SSILENT=1
      SILENT=1
      VERBOSE=0
    ;;
    --mk-conf)
      mkConf
    ;;
    -H|--host-mode)
      HOSTMODE=1
    ;;
    -C|--critical-error)
      CRITICALERROR=1
    ;;
    -E|--exit-error)
      EXITERROR=1
    ;;
    -S|--start-host)
      STARTHOST="$2"

      shift
    ;;
    -n|--nr-pad)
      PAD="$2d"

      shift
    ;;
    -p|--pad)
      ENABLEPAD=1
    ;;
    -P|--password)
      SSHPASSWD="$2"

      shift
    ;;
    -i|--start-ip)
      START="$2"

      shift
    ;;
    -t|--stop-ip)
      NUMHOSTS="$2"

      shift
    ;;
    -k|--keys)
      SSHPASSWD=""
      KEYSONLY=1
    ;;
    -F|--file)
      FILE="$2"
      FILEMODE=1

      shift
    ;;
    -U|--sudo)
      USESUDO=1
    ;;
    --no-sudo)
      USESUDO=0
    ;;
    -*)
      printEE "Unknown switch '$1'."
    ;;
    *)
      VAL[$GO]="$1"
      GO=$((GO+1))
    ;;
  esac

  shift
done

printIC "Checking so that we have all args needed."

if [ -z "$START" ];
then
  if [ $FILEMODE -eq 1 ];
  then
    printIC "Ignoring missing start ip, since we are in file mode."
  elif [ -n "${VAL[0]}" ];
  then
    START="${VAL[0]}"
  else
    printEE "Missing start ip. See $0 -h, for help."
  fi
fi

if [ -z "$NUMHOSTS" ];
then
  if [ $FILEMODE -eq 1 ];
  then
    printIC "Ignoring missing stop ip, since we are in file mode."
  elif [ -n "${VAL[1]}" ];
  then
    NUMHOSTS="${VAL[1]}"
  else
    printEE "Missing stop ip. See $0 -h, for help."
  fi
fi

if [ $KEYSONLY -eq 1 ] && [ "$USESUDO" == "1" ];
then
  printEE "You cannot use ssh-keys with the sudo function."
fi

if [ $HOSTMODE -eq 0 ] && [ $FILEMODE -eq 0 ];
then
  ## Check if we got an IP-address.

  if [ -z "`echo "$START" | grep -sE "^(.*?@)?([1-2]?[0-9]?[0-9])\.([1-2]?[0-9]?[0-9])\.([1-2]?[0-9]?[0-9])\.([1-2]?[0-9]?[0-9])$"`" ];
  then
    printEE "This does not look like an IP-address: $START."
  fi

  INITIAL="`basename \`echo "$START" | sed 's/\./\//g'\``"
  HOSTSEQ="`echo "$START" | sed 's/\./ /g' | awk '{ print $1 "." $2 "." $3 "." }'`"
elif [ $HOSTMODE -eq 1 ] && [ $FILEMODE -eq 1 ];
then
  printEE "Host mode and file mode is not compatible."
elif [ $FILEMODE -eq 1 ];
then
  printIC "No IP/host name checks are done since we are in filemode."
else
  if [ -n "`echo "$START" | grep -sE "^(.*?@)?([1-2]?[0-9]?[0-9])\.([1-2]?[0-9]?[0-9])\.([1-2]?[0-9]?[0-9])\.([1-2]?[0-9]?[0-9])$"`" ];
  then
    printWC "This looks like an IP: $START, but hostmode is set?"
  fi
fi

if [ "$SSHPASSWD" == "" ] && [ $KEYSONLY -eq 0 ];
then
  echo -n "Password or ssh-key? [P|k]: "
  read -n 1 ANSW
  echo

  if [ "$ANSW" == "p" ] || [ "$ANSW" == "P" ] || [ "$ANSW" == "" ] && [ -z "$SSHPASSWD" ];
  then
    echo -n "Password: "

    read -s SSHPASSWD
    echo

    if [ -z "$SSHPASSWD" ];
    then
      printEE "No password. Bye."
    fi
  elif [ "$ANSW" != "k" ] && [ "$ANSW" != "K" ] && [ -z "$SSHPASSWD" ];
  then
    printEE "What? No fooling around!"
  fi
fi

if [ $FILEMODE -eq 0 ];
then
  printIC "Start: $START. Stop: $NUMHOSTS."
fi

if [ $SILENT -eq 1 ];
then
  SSHOPTS="-q"
  EXPOPTS="1"
elif [ $VERBOSE -eq 1 ];
then
  SSHOPTS="-v"
else
  EXPOPTS=""
  SSHOPTS=""
fi

LASTRETVAL=0
PERMFAIL=0

if [ $HOSTMODE -eq 0 ] && [ $FILEMODE -eq 0 ];
then
  printIC "Using IP mode."

  for i in `seq $INITIAL $NUMHOSTS`
  do
    ipConnect "$i"
  done
elif [ $FILEMODE -eq 1 ];
then
  printIC "Using file mode."

  ## Check file perm
  printIC "Check file readability."

  if [ ! -r "$FILE" ];
  then
    printEE "In filemode, but cannot read file: $FILE."
  fi

  ## Read file
  HOSTS="`cat "$FILE" 2>/dev/null || echo "FAILED"`"

  if [ "$HOSTS" == "FAILED" ];
  then
    printEE "Failure to read contents of file: $FILE."
  fi

  for HOST in $HOSTS
  do
    ipConnect "$HOST"
  done
else
  printIC "Using hostmode."

  ## Hostmode

  for i in `seq $STARTHOST $NUMHOSTS`
  do
    hostConnect "$i"
  done
fi

printWC "------------------------------------------------------------" 1
printWC "## I: Done!" 1

if [ $PERMFAIL -eq 1 ];
then
  exit 1
else
  exit $LASTRETVAL
fi
