#!/usr/bin/env python

## Part of Arxx SSH Tools. Published at: <https://arxx.se>.
## Written by: Johan Hyllstedt. You can reach me at: <code@arxx.se>.
## See README for more information.

## Copyright 2010, 2011, 2012 Johan Hyllstedt.

## This file is part of Arxx SSH Tools.

## Arxx SSH Tools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## Arxx SSH Tools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with Arxx SSH Tools.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import string
import readline
import shutil
import errno

## Constants
EXEC = 1
NOEXEC = 2

## Variables
INTERACTIVE = False ## Global
## List of files to inst/uninst
files = [
          [ 'Astgui.glade', NOEXEC ],
          [ 'HISTORY', NOEXEC ],
          [ 'README', NOEXEC ],
          [ 'SshLoginPasswdInteract.exp', EXEC ],
          [ 'ScpLoginPasswdPull.exp', EXEC ],
          [ 'ScpSinglePasswd.exp', EXEC ],
          [ 'ScpLoginPasswdPush.exp', EXEC ],
          [ 'SshCmdLoginPasswd.exp', EXEC ],
          [ 'WrapperScp.sh', EXEC ],
          [ 'WrapperSsh.sh', EXEC ],
          [ 'astgui', EXEC ],
          [ 'astcmd', EXEC ],
          [ 'astgo', EXEC ],
          [ 'astlogin', EXEC ],
          [ 'astscp', EXEC ],
          [ 'SshCmdLoginPasswdSudo.exp', EXEC ],
          [ 'SshLoginPasswdInteractSudo.exp', EXEC ],
          [ 'CmdSinglePasswd.exp', EXEC ],
          [ 'COPYING', NOEXEC ]
        ]
wrappers = { 'scp' : 'WrapperScp.sh', 'ssh' : 'WrapperSsh.sh' }
symlinks = [
             [ 'scpMultiSSHTool.sh', 'astscp' ],
             [ 'astgui.py', 'astgui' ],
             [ 'cmdMultiSSHTool.sh', 'astcmd' ],
             [ 'loginMultiSSHTool.sh', 'astlogin' ] ## Symlinks are not moved any more due to a shutil bug, now we make them. Thus format link name, target.
           ]
mainbins = [ 'astcmd', 'astgui', 'astlogin', 'astscp', 'astgo' ] ## Main bins can be symlinked in to /usr/bin by request
home = os.getenv("HOME")
publicPath = "/opt/sshTools/" ## Global path for root-mode
defaultPath = home + "/sshTools/" ## Userspace path
globalCfg = "/etc/ast.cfg"
usrbin = "/usr/bin/" ## Can be /usr/local/bin if you change here ;-)
kupfer_user_path = home + "/.local/share/kupfer/plugins/" ## If we detect a kupfer installation install to these locations
kupfer_global_path = "/usr/share/kupfer/kupfer/plugin/"
profile_d = '/etc/profile.d/ast.sh'
profile = '/etc/profile'

def inst_wrappers(installpath, glob = False):
  if (installpath == '' or installpath == '//') and not glob:
    installpath = pathFix(os.path.abspath(os.path.dirname(__file__)))
  elif (installpath == '' or installpath == '//') and glob:
    installpath = publicPath

  chmod = False

  if glob:
    if os.path.exists('/etc/profile.d/'):
      if os.path.exists(profile_d):
        print "I: We already seem to have setup the wrappers! (%s)" %profile_d
        return
      else:
        file = profile_d
        chmod = True
        mode = 0644
    else:
      ## Check if we've already written to this!
      file = profile
      if not os.path.exists(profile):
        print "W: You do not seem to have a global profile, this might not work very well! (%s)"%profile
        chmod = True
        mode = 0644
      else:
        try:
          with open(file, 'r') as some_file:
            data = some_file.read()
        except:
          print "W: Failed to read global profile, cannot determine if we've already written to it! (%s)"%file
        else:
          if data.find('Below lines inserted by Arxx SSH Tools Installer') > -1:
            ## No need to write to it again!
            print "W: Already written to global profile, not doing it again! (%s)"%file
            return

    text = "\n## Below lines inserted by Arxx SSH Tools Installer.\nalias ssh='%s'"%pathFix(installpath + '/' + wrappers['ssh'], True)
    text += "\nalias scp='%s'\n"%pathFix(installpath + '/' + wrappers['scp'], True)

    try:
      f = open(file, 'a')
      f.write(str(text))
    except:
      print "E: Failed to append lines to: %s." % file
      sys.exit(1)
    else:
      f.close()

    if chmod:
      try:
        os.chmod(file, mode)
      except:
        print "W: Failed to alter file permissions on: %s to: %s." % (file, mode)
  else:
    file = os.getenv("HOME") + "/.bashrc"

    if not os.path.exists(file):
      print "W: You do not seem to have a bashrc file? (%s)" % file
    else:
      ## Check if we've already written to this!
      try:
        with open(file, 'r') as bashrc_file:
          data = bashrc_file.read()
      except:
        print "W: Failed to read bashrc file, cannot determine if we've already written to it! (%s)"%file
      else:
        if data.find('Below lines inserted by Arxx SSH Tools Installer') > -1:
          ## No need to write to it again!
          print "W: Already written to bashrc file, not doing it again! (%s)"%file
          return

    text = "\n## Below lines inserted by Arxx SSH Tools Installer.\nalias ssh='%s'"%pathFix(installpath + '/' + wrappers['ssh'], True)
    text += "\nalias scp='%s'\n"%pathFix(installpath + '/' + wrappers['scp'], True)

    try:
      f = open(file, 'a')
      f.write(str(text))
    except:
      print "E: Failed to append lines to: %s." % file
      sys.exit(1)
    else:
      f.close()

def interactive(uninst = False):
  print "I: Entering interactive-mode."
  print "I: Untill you begin (un)installing, you can cancel at anytime. (Type 'exit')"
  global INTERACTIVE
  INTERACTIVE = True

  ## What do we want answered?
  #mode = 0 ## 1 = Inst, 2 = Uinst
  location = "" ## "" = default
  useCurdir = False ## Use current dir?
  public = False
  symlink = False
  inst_wrapp = False

  if not uninst:
    answg, answb, answe = question('Q: Install or Uninstall? [I/u]: ', [ '', 'i', 'install' ], [ 'u', 'uninstall'], True)
  else:
    answg, answb, answe = question('Q: Install or Uninstall? [i/U]: ', [ 'i', 'install' ], [ '', 'u', 'uninstall'], True)

  if answg:
    print "I: Install-mode."

    check()

    answg, answb, answe = question('Q: Is the directory the files is in now OK? [y/N]: ', [ 'y', 'yes' ], [ '', 'n', 'no'], True)

    if answg:
      useCurdir = True
    elif answb:
      location = question('Q: Location to install to? (empty for default): ', [ '' ], None, True, True)

      if location != "":
        ## Use speicifed, check if exists
        if not os.path.exists(location):
          print "I: Path does not exist already: %s." %location

          answg, answb, answe = question('Q: Create this directory? [Y/n]: ', [ '', 'y', 'yes' ], [ 'n', 'no'], True)

          if not answg:
            print "E: Ok, bye then."
            sys.exit(0)
          else:
            try:
              os.makedirs(location, 0700)
            except OSError as excep:
              if excep.errno == errno.EEXIST:
                pass
              else:
                errorDisp("E: Failed to make dir: " + location + ".")
                sys.exit(1)

    answg, answb, answe = question('Q: Will this folder be used by others than you (system-wide install)? [y/N]: ', [ 'y', 'yes' ], [ '', 'n', 'no'], True)

    if answg:
      if os.getuid() != 0:
        print "W: As you are not installing as root, you will need to make sure the owner is correct after installing."
        print "I: The usual owner of system-wide files are root, as you do not want a regular user to be able to update files used by everyone."

      public = True

      answg, answb, answe = question('Q: Do you also wish to create symlinks in ' + usrbin + '? [Y/n]: ', [ '', 'y', 'yes' ], [ 'n', 'no'], True)

      if answg:
        symlink = True

    answg, answb, answe = question('Q: Do you want to use the Wrapper scripts? [y/N]: ', [ 'y', 'yes' ], [ '', 'n', 'no'], True)

    if answg:
      inst_wrapp = True

    ## Carry out install
    answg, answb, answe = question('Q: The installation will begin now, is that OK? [Y/n]: ', [ '', 'y', 'yes' ], [ 'n', 'no'], True)

    if answb:
      print "E: Ok, bye then."
      sys.exit(0)

    install(location, useCurdir, public, symlink, inst_wrapp)
  elif answb:
    print "I: Uninstall-mode."

    location = question('Q: Where did you install it? (empty for default): ', [ '' ], None, True, True)

    answg, answb, answe = question('Q: Should the cfg file be deleted too? [y/N]: ', [ 'y', 'yes' ], [ '', 'n', 'no'], True)

    delcfg = answg ## As we ask another q...

    answg, answb, answe = question('Q: Ready to uninstall, proceed? [Y/n]: ', [ '', 'y', 'yes' ], [ 'n', 'no'], True)

    if answb:
      print "E: Ok, bye then."
      sys.exit(0)

    if delcfg:
      uninstall(location, True)
    else:
      uninstall(location, False)

  sys.exit(0)

def question(quest, goodv, badv, exitOk = False, returnAnsw = False):
  ## Order return: good, bad, exit

  ## Asks the question, looking for good or bad or exit and returns True or False if it matches.
  ## All answers are case in-sensitive.

  ## If exitOk is true, it will exit without calling back.

  cont = True
  good = False
  bad = False
  exit = False

  while cont:
    try:
      answ = str(raw_input(quest))
    except EOFError:
      if exitOk:
        print "\nI: Bye."
        sys.exit(0)
      else:
        exit = True
        break
    except KeyboardInterrupt:
      if exitOk:
        print "\nI: Bye."
        sys.exit(0)
      else:
        exit = True
        break

    if returnAnsw:
      return answ

    for g in goodv:
      if answ.lower() == str(g).lower():
        good = True
        cont = False
        break

    for b in badv:
      if answ.lower() == str(b).lower():
        bad = True
        cont = False
        break

    if answ.lower() == 'quit' or answ.lower() == 'exit' or answ.lower() == 'bye' or answ.lower() == 'q':
      exit = True

      if exitOk:
        print "I: Bye."
        sys.exit(0)

      break

    if cont:
      print "I: What?! Type 'exit' to quit."

  return good, bad, exit

def unlinkDir(dirn):
  dirn = pathFix(dirn)

  print "I: Trying to delete: %s." %dirn

  try:
    shutil.rmtree(dirn)
  except:
    errorDisp("Failed to remove dir tree: " + dirn + ".")
    sys.exit(1)
  else:
    print "I: Done."

def validateDir(dirn, picky = False):
  ## Not picky validater, only one file match = Match
  ## Use picky to be picky :-), in this mode all files must be there.

  global files

  match = False
  permfail = False

  for filename, filemode in files:
    if os.path.exists(os.path.join(dirn, filename)):
      match = True
    else:
      if picky:
        permfail = True

  if permfail:
    match = False ## Always fail on permfail

  return match

def pathFix(pathn, isFile = False):
  if isFile:
    return str(os.path.normpath(pathn))
  else:
    return str(os.path.normpath(pathn) + '/')

def uninstall(path, cfgToo = False):
  print "I: Uninstalling."

  if path != "":
    path = pathFix(path) ## Otherise it will try to uninstall from base dir...

  ## Guess work
  testPaths = [ path, pathFix(defaultPath), pathFix(publicPath) ]
  uninst = False
  global INTERACTIVE

  for path in testPaths:
    if os.path.exists(path):
      ## Then we have made a correct choice perhaps
      if validateDir(path):
        if INTERACTIVE:
          print "I: Found match in: '%s'." %path
          answg, answb, answe = question('Q: This directory looks like you have installed to, OK to delete? [y/N]: ', [ 'y', 'yes' ], [ '', 'n', 'no'], True)

          if answg:
            unlinkDir(path)
            uninst = True
            break
        else:
          unlinkDir(path)
          uninst = True
          break
      else:
        print ("I: I was given the path: " + path + " but no installation was found there.")

  if not uninst:
    print "W: Did not uninstall anything."

  ## Unlink cfg
  if cfgToo:
    if not delCfg():
      print "W: Failed to delete cfg file."
    else:
      print "I: Cfg deleted."

  ## If symlinks exists, delete them too
  for thisbin in mainbins:
    symdest = usrbin + thisbin

    if os.path.lexists(symdest):
      try:
        os.unlink(symdest)
      except:
        print "W: Failed to delete symlink %s." % symdest

  if os.path.exists(profile_d):
    try:
      os.unlink(profile_d)
    except:
      print "W: Failed to remove file: %s." % profile_d

  print "-------------------------"
  print "If you chose to use the wrapper scripts, do not forget to manually clean up (if needed):"
  print "~/.bashrc or /etc/profile"
  print "Look for the line: 'Below lines inserted by Arxx SSH Tools Installer'."
  print "-------------------------"

  print "Uninstall done!"

  sys.exit(0)

def install(newpath, useCurdir = False, public = False, symlinkIn = False, inst_wrapp = False):
  curdir = pathFix(os.path.abspath(os.path.dirname(__file__)))
  kupfer = False

  if not glob_kupfer_force:
    kupfer = check(['kupfer'], False, False)
  else:
    kupfer = True

  if newpath == "" and not useCurdir:
    ## Use default
    newpath = getDefaultPath()
    public = getIsPublic(newpath)

    if public:
      symlinkIn = True
  else:
    newpath += '/'

  newpath = pathFix(newpath)

  if newpath == publicPath or public:
    newmode = 0755
    newmodeNE = 0644
  else:
    newmode = 0700
    newmodeNE = 0600

  if not validateDir(curdir, True):
    print "E: Failed to determine full path to this dir (missing files in source?)..."
    sys.exit(0)

  if kupfer:
    print "I: Will try to install kupfer plugin."

    if public:
      kupfer_dir = kupfer_global_path
    else:
      kupfer_dir = kupfer_user_path

    if not os.path.exists(kupfer_dir):
      try:
        os.makedirs(kupfer_dir, newmode)
      except OSError as excep:
        if excep.errno == errno.EEXIST:
          pass
        else:
          errorDisp("W: Failed to make dir (for kupfer plugin): " + kupfer_dir + ".")
          kupfer = False

      if public:
        try:
          os.chown(kupfer_dir, 0, 0)
        except:
          print "W: Failed to alter file ownership on: %s, to root." % (kupfer_dir)


    if kupfer:
      print "I: Moving kupfer plugin to kupfer plugins dir: '%s'."%kupfer_dir
      cFile = 'kupferplugin_astgo.py'
      src = str(curdir + "/" + cFile)
      dest = str(kupfer_dir + cFile)

      if os.path.exists(src) and not os.path.exists(dest):
        try:
          shutil.move(src, dest)
        except:
          print "W: Failed to relocate file to: %s." %dest

        try:
          os.chmod(dest, newmodeNE)
        except:
          print "W: Failed to alter file permissions on: %s to: %s." % (dest, newmodeNE)

        if public:
          try:
            os.chown(dest, 0, 0)
          except:
            print "W: Failed to alter file ownership on: %s, to root." % (dest)
      else:
        print "W: File not found %s or %s already exists!" % (src, dest)
  else:
    print "I: Not installing kupfer plugin, use -K, to force install!"

  if not useCurdir:
    print "I: Not set to use current dir, so trying to install to: %s." %newpath

    if os.path.exists(newpath) and validateDir(newpath):
      print "W: The path: %s already exists and contains an installation, assuming that you just want to write a cfg file instead!" %newpath
      useCurdir = True
    elif not os.path.exists(newpath):
      print "I: Making new home: %s." %newpath

      try:
        os.makedirs(newpath, 0700)
      except OSError as excep:
        if excep.errno == errno.EEXIST:
          pass
        else:
          print "E: Failed to make dir: %s." %newpath
          sys.exit(1)

    if not useCurdir:
      for cFile, cMode in files:
        src = str(curdir + "/" + cFile)
        dest = str(newpath + cFile)

        print "I: Moving '%s' to new home." % src

        if os.path.exists(src) and not os.path.exists(dest):
          try:
            shutil.move(src, dest)
          except:
            print "E: Failed to relocate file to: %s." %dest
            sys.exit(1)

          if cMode == EXEC:
            destmode = newmode
          elif cMode == NOEXEC:
            destmode = newmodeNE

          try:
            os.chmod(dest, destmode)
          except:
            print "E: Failed to alter file permissions on: %s to: %s." % (dest, destmode)
            sys.exit(1)

          if public:
            try:
              os.chown(dest, 0, 0)
            except:
              print "W: Failed to alter file ownership on: %s, to root." % (dest)
        else:
          print "E: File not found %s or %s already exists!" % (src, dest)
          sys.exit(0)

      for linkname, linktarget in symlinks:
        linkname = pathFix(newpath + linkname, True)
        linktarget = pathFix(newpath + linktarget, True)

        print "I: Symlinking '%s'." % linkname

        if os.path.lexists(linktarget) and not os.path.lexists(linkname):
          try:
            os.symlink(linktarget, linkname)
          except:
            print "E: Failed to symlink: %s to: %s." % (linkname, linktarget)
            sys.exit(1)
        else:
          print "E: File not found %s or %s already exists!" % (linktarget, linkname)
          sys.exit(1)

  if symlinkIn:
    ## Create symlinks in /usr/bin to main bins

    for thisbin in mainbins:
      if useCurdir:
        symsrc = curdir + thisbin
      else:
        symsrc = newpath + thisbin

      symdest = usrbin + thisbin

      try:
        os.symlink(symsrc, symdest)
      except:
        print "E: Failed to symlink '%s' to '%s'." % (symdest, symsrc)
        sys.exit(1)
      else:
        print "I: Made symlink '%s' to '%s'." % (symdest, symsrc)

  ## Chmod ast dir too
  if not useCurdir:
    astdir = os.path.dirname(newpath)
  else:
    astdir = os.path.dirname(curdir + '/')

  try:
    os.chmod(astdir, newmode)
  except:
    print "E: Failed to change permissions on '%s' to '%s'." % (astdir, newmode)
    sys.exit(1)

  if public:
    try:
      os.chown(astdir, 0, 0)
    except:
      print "W: Failed to alter file ownership on: %s, to root." % (dest)

  ## Write cfg file for ssh tools
  if useCurdir:
    cfg = "## Autogenerated cfg by Arxx SSH Tools Installer.\nSCPATH=" + curdir
  else:
    cfg = "## Autogenerated cfg by Arxx SSH Tools Installer.\nSCPATH=" + newpath

  if public:
    writeCfg(cfg, True)
  else:
    writeCfg(cfg)

  if inst_wrapp:
    inst_wrappers(newpath, public)

  print "I: Install succesfull!"

def check(binaries = [ 'bash', 'expect', 'awk', 'sed', 'grep', 'python' ], exit_ok = True, verbose = True):
  ## Search for paths, look for dependencies
  paths = os.getenv("PATH")

  if verbose:
    print "I: Searching for requirements."

  for binary in binaries:
    if not searchFile(str(binary), paths):
      if exit_ok:
        print "E: Cannot locate required binary: '%s'." % binary
        sys.exit(1)
      else:
        return False

  if verbose:
    print "I: All requirements met."

  return True

def delCfg():
  cfgFile = os.getenv("HOME") + "/.sshTools.cfg"
  status = False

  if os.path.exists(globalCfg):
    try:
      os.unlink(globalCfg)
    except:
      status = False
    else:
      status = True

  if os.path.exists(cfgFile):
    try:
      os.unlink(cfgFile)
    except:
      status = False
    else:
      status = True

  return status

def writeCfg(text, glob = False):
  if not glob:
    cfgFile = os.getenv("HOME") + "/.sshTools.cfg"
    newmode = 0600
  else:
    cfgFile = globalCfg
    newmode = 0644

  if os.path.exists(cfgFile):
    print "W: A cfg file exists already. Delete: '%s' first if you wish to make a new one!" % cfgFile
  else:
    try:
      f = open(cfgFile, 'w')
      f.write(str(text))
    except:
      print "E: Failed to write new cfg to: %s." % cfgFile
      sys.exit(1)
    else:
      f.close()

      try:
        os.chmod(cfgFile, newmode)
      except:
        print "W: Failed to change permissions on cfg file."
      else:
        print "I: Succesfully wrote new cfg file to '%s'." % cfgFile

      if glob:
        try:
          os.chown(cfgFile, 0, 0)
        except:
          print "W: Failed to alter file ownership on: %s, to root." % (cfgFile)

def errorDisp(text):
  ## Prints to stderr
  print >>sys.stderr, text

def dispHelp():
  print "  This installer has few options. Usually just running it will be ok!"
  print "  NOTE: If you want full control you should read the README file and set things up manually, it is not hard!"
  print ""
  print "  The default is to try and create the dir ~/sshTools and set things up there, if run as a regular user. However if,"
  print "  you are root, it will try to install to /opt/sshTools and create symlinks in %s. Unless the argument -C or the above dir(s) already exists" %usrbin
  print "  and has a previous installation, it then tries to set things up using the dir this script is from."
  print "  You can also use -d and it will use that directory to set things up in."
  print "  There is also -I, for interactive mode or -u/i for (un)install."
  print ""
  print "  Full Commandline Syntax:"
  print "    -h              - This help."
  print "    -C              - Use current directory to install to."
  print "    -d              - Use alternative directory to install/uninstall to."
  print "    -I              - Enter interactive-mode."
  print "    -u              - Uninstall, this works with -d if you have set it up differently than default."
  print "    -f              - When uninstalling also delete cfg file."
  print "    -i              - Install."
  print "    -G              - This is a global/system-wide install."
  print "    -S              - Create symlinks to binaries in %s." %usrbin
  print "    -K              - Force installation of kupfer plugin."
  print "    -W              - Setup the wrapper scripts."
  print ""
  print "  FRIENDLY NOTE: The default mode is now to enter interactive mode, unless the options -i or -u are used."

  sys.exit(0)

def searchFile(filename, search_paths):
  paths = string.split(search_paths, ':')

  for path in paths:
    if os.path.exists(os.path.join(path, filename)):
      return True

def getDefaultPath():
  newpath = pathFix(defaultPath)

  ## Check if root, set path then
  if os.getuid() == 0:
    newpath = publicPath

  return newpath

def getIsPublic(path):
  if path == publicPath:
    return True
  else:
    return False

if __name__ == "__main__":
  newpath = getDefaultPath()
  useCurdir = False
  uninst = False
  inst = False
  cfgDel = False
  public = False
  symlink = False
  uninstTog = False
  inst_wrapp = False

  global glob_kupfer_force
  glob_kupfer_force = False

  try:
    if os.path.basename(sys.argv[0]) == 'uninstall.py':
      uninst = True
  except:
    pass

  ## Read args
  argnr = 0
  for inputparam in sys.argv:
    twice = False

    if inputparam == "-h":
      dispHelp()
    elif inputparam == "-C":
      useCurdir = True
    elif inputparam == "-W":
      inst_wrapp = True
    elif inputparam == "-G":
      public = True
    elif inputparam == "-S":
      symlink = True
    elif inputparam == "-f":
      cfgDel = True
    elif inputparam == "-K":
      glob_kupfer_force = True
    elif inputparam == "-d":
      ## Switch def dir
      try:
        newpath = sys.argv[argnr + 1]
      except:
        errorDisp("E: Failed to get alternate dir.")
        sys.exit(1)
      else:
        twice = True
    elif inputparam == "-I":
      interactive()
    elif inputparam == "-u":
      uninst = True ## Do like this so we can get -d after
      uninstTog = True
    elif inputparam == "-i":
      inst = True

    if twice:
      argnr += 2
    else:
      argnr += 1

  os.umask(0)

  if uninst and uninstTog:
    uninstall(newpath, cfgDel)
  elif inst:
    check()
    install(newpath, useCurdir, public, symlink, inst_wrapp)
  else:
    interactive(uninst)
